import dataclasses
from typing import Literal, Union

import numpy as np

from .synth_format import RailFilter, DataContainer, WALLS, WALL_LOOKUP, WALL_SYMMETRY, WALL_TYPES, NOTE_TYPES
from .utils import bounded_arange
from . import movement, rails

# generic helpers
def angle_to_xy(angles: "numpy array like") -> "numpy array like":
    """convert a angles in degrees to x and y coordinates"""
    rad_angles = np.radians(angles)
    return np.stack((np.cos(rad_angles), np.sin(rad_angles)), -1)

def random_ring(count: int) -> "numpy array (n, 2)":
    """random positons along a ring of radius 1"""
    return angle_to_xy(np.random.random_sample((count,)) * 360)

def random_xy(count: int, min: "numpy array (2)", max: "numpy array (2)") -> "numpy array (n, 2)":
    """random positions in a rectangle"""
    return np.random.random_sample((count, 2)) * (max - min) + min

# pattern generation
def spiral(
    fidelity: float, length: float, start_angle: float = 0
) -> "numpy array (n, 2)":
    """spiral with radius 1, uses random when fidelity is 0"""
    if not fidelity: 
        return random_ring(int(length))
    rot = np.arange(length) / fidelity  # in rotations, ie 1.0 = 360°
    if length % 1:
        # add a partial angle at the end
        rot = np.concatenate((rot, np.array([(length / fidelity)])))
    return angle_to_xy(rot * 360 + start_angle)

def add_spiral(nodes: "numpy array (n, 3)", fidelity: float, radius: float, start_angle: float = 0.0, direction: int = 1) -> "numpy array (n, 3)":
    nodes = nodes.copy()
    nodes[:, :2] += spiral(
        fidelity=fidelity * direction,
        length=nodes.shape[0],
        start_angle=start_angle if direction == 1 else 180 - start_angle
    ) * radius
    return nodes

def spikes(
    fidelity: float, length: float, start_angle: float = 0
) -> "numpy array (n, 2)":
    """spikes with radius 1, uses random when fidelity is 0"""
    output = np.zeros((int(length) * 3, 2))
    output[1::3] = (
        spiral(fidelity, length, start_angle) 
    )
    return output

def add_spikes(nodes: "numpy array (n, 3)", fidelity: float, radius: float, spike_duration: float, start_angle: float = 0.0, direction: int = 1) -> "numpy array (n, 3)":
    node_count = nodes.shape[0]  # backup count before repeat
    nodes = np.repeat(nodes, 3, axis=0)
    nodes[::3, 2] -= spike_duration
    nodes[1::3, 2] -= spike_duration/2
    nodes[:, :2] += spikes(
        fidelity=fidelity * direction,
        length=node_count,
        start_angle=start_angle if direction == 1 else 180 - start_angle
    ) * radius
    return nodes

def create_lightning(data: DataContainer, interval: float, rail_filter: RailFilter|None=None) -> None:
    """create a zigzag between left hand and right hand notes and rails

    It picks the first rail (or longer if starting simultaneous) from either left and right as primary.
    Then it interpolates that, picking every other node from the secondary color if there is a rail/note there
    """
    new_filter = RailFilter(single=False) if rail_filter is None else dataclasses.replace(rail_filter, single=False)
    d = data.filtered(rail_filter=new_filter)
    if not d.left:
        raise ValueError("Input does not contain any left rails")
    if not d.right:
        raise ValueError("Input does not contain any right rails")
    first_l = sorted(d.left)[0]
    first_r = sorted(d.right)[0]
    end_l = d.left[first_l][-1,2]
    end_r = d.right[first_r][-1,2]

    primary = "right"
    if first_l < first_r or ((first_l == first_r) and (end_l > end_r)):
        # use left as primary if it starts first OR starts at the same time and ends later
        primary = "left"
    primary_nodes = d.left[first_l] if primary == "left" else d.right[first_r]
    secondary_notes = d.right if primary == "left" else d.left

    out = rails.interpolate_nodes(primary_nodes, "spline", interval=interval)
    for i in range(1, len(out), 2):
        new_pos = rails.get_position_at(secondary_notes, out[i,2], interpolate_gaps=False)
        if new_pos is not None:
            out[i, :2] = new_pos
    data.both = data.both | {out[0,2]: out}

def create_parallel(data: DataContainer, distance: float, types: tuple[str, ...] = NOTE_TYPES, rail_filter: RailFilter|None=None) -> None:
    """create parallel patterns by splitting specials, or adding the other hand
    
    when splitting specials, both will be moved by half the distance left and right.
    otherwise, the existing color stays, while the new hand is moved by distance in the "natural" direction.

    distance can be negative to create crossovers.
    """
    left_orig, right_orig = data.left, data.right  # create a backup of the input
    data.left = (
        left_orig
        | ({} if "right" not in types else {t: nodes - [distance,0,0] for t, nodes in sorted(right_orig.items()) if not rail_filter or rail_filter.matches(nodes)})  # shift over right hand by <distance>
        | ({} if "single" not in types else {t: nodes - [distance/2,0,0] for t, nodes in sorted(data.single.items()) if not rail_filter or rail_filter.matches(nodes)})  # shift over single & both by <distance>/2
        | ({} if "both" not in types else {t: nodes - [distance/2,0,0] for t, nodes in sorted(data.both.items()) if not rail_filter or rail_filter.matches(nodes)})
    )
    # vice versa for right hand
    data.right = (
        right_orig 
        | ({} if "left" not in types else {t: nodes + [distance,0,0] for t, nodes in sorted(left_orig.items()) if not rail_filter or rail_filter.matches(nodes)})
        | ({} if "single" not in types else {t: nodes + [distance/2,0,0] for t, nodes in sorted(data.single.items()) if not rail_filter or rail_filter.matches(nodes)})
        | ({} if "both" not in types else {t: nodes + [distance/2,0,0] for t, nodes in sorted(data.both.items()) if not rail_filter or rail_filter.matches(nodes)})
    )
    # wipe single & both (leaving only those NOT affected)
    if "single" in types:
        data.single = {} if not rail_filter else {t: nodes for t, nodes in sorted(data.single.items()) if not rail_filter.matches(nodes)}
    if "both" in types:
        data.both = {} if not rail_filter else {t: nodes for t, nodes in sorted(data.both.items()) if not rail_filter.matches(nodes)}

def find_wall_patterns(walls: WALLS) -> list[tuple[int, int, float]]:
    """try to find repeating "patterns" with identical wall type and timing, returns wall count per pattern, pattern count and pattern length for all possible candidates"""
    wall_count = len(walls)
    if wall_count < 2:
        raise ValueError("Need at least two walls to find repeating patterns")
    wall_types = [w[0,3] for _, w in sorted(walls.items())]

    # find candidates for patterns, by looking where the first wall type is repeated
    matching_first = [i for i, t in enumerate(wall_types) if i and t == wall_types[0]]
    if not matching_first:
        raise ValueError(f"Could not find any repeats of first wall ({WALL_LOOKUP[wall_types[0]]})")
    # now ensure it is an divisor of the wall count
    matching_count = [i for i in matching_first if wall_count % i == 0]
    if not matching_count:
        raise ValueError(f"Could not find any repeats of first wall that are divisors of total wall count ({wall_count})")
    # now check if types match
    matching_types = [i for i in matching_count if wall_types == wall_types[:i] * (wall_count//i)]
    if not matching_types:
        raise ValueError(f"Could not find any wall type pattern that repeats consistently")
    # finally check times
    wall_times = np.array([w[0,2] for _, w in sorted(walls.items())])
    out = []
    for walls_per_pattern in reversed(matching_types):  # start with the largest candidate
        reshaped = wall_times.reshape((-1, walls_per_pattern)).copy()
        reshaped -= reshaped[:,0,np.newaxis]
        if np.allclose(reshaped[0], reshaped):
            out.append((walls_per_pattern, wall_count//walls_per_pattern, reshaped[0,-1]-reshaped[0,0]))
    if out:
        return out
    raise ValueError(f"Could not find any pattern of type AND timing that repeats consistently")

def blend_wall_single(first: "numpy array (n, 5)", second: "numpy array (n, 5)", interval: float, with_symmetry: bool=True) -> dict[float, "numpy array (1, 5)"]:
    """create blending substeps between two patterns
    
    first and second must be numpy arrays of the patterns (each with n walls)
    with_symmetry can be set to false to disregard that symetrical walls (like squares) look the same in multiple orientations
    """
    delta_t = second[0,2] - first[0,2]
    if not np.allclose(first[:,3], second[:,3]):
        raise ValueError("Patterns have mismatched wall types")
    if not np.allclose(first[:,2], second[:,2] - delta_t):
        raise ValueError("Patterns have mismatched timing")
    time_offsets = bounded_arange(0, delta_t, interval)

    out_walls: dict[float, "numpy array (1, 5)"] = {}

    for f, s in zip(first, second):
        sym = WALL_SYMMETRY[WALL_LOOKUP[f[3]]] if with_symmetry else 360
        f_ang = f[4] % sym
        s_ang = s[4] % sym

        if f_ang == s_ang:  # equal angle => shift
            delta_xyttr = np.zeros_like(f) 
            delta_xyttr[:2] = (s[:2] - f[:2]) / delta_t
            delta_xyttr[2] = 1
            # x and y: delta for t=1, t: 1, type and rotation: 0
            for i in time_offsets:
                new = (f + delta_xyttr * i)
                out_walls[new[2]] = new[np.newaxis]
        else:  # spiral stack logic
            ang = ((s_ang - f_ang)+sym/2)%sym-sym/2  # angle difference, between -(sym/2) and +(sym/2), ie -180 to 180 for non-symetrical walls
            piv = f[:3] + movement.rotate((s - f)[:3]/2, 90 - ang/2) / np.sin(np.radians(ang/2))  # pivot location matching rotation with translation

            for i in time_offsets:
                new = movement.rotate(f[np.newaxis], angle=ang * (i/delta_t), pivot=piv)
                new[:,2] += i
                out_walls[new[0,2]] = new
    return out_walls

def blend_walls_multiple(patterns: "numpy array (m,n,5)", interval: float, with_symmetry: bool=True) -> dict[float, "numpy array (1, 5)"]:
    """calls blend_wall_single to blending between multiple patterns, each with n walls"""
    out_walls: dict[float, "numpy array (x*n, 5)"] = {}
    for first, second in zip(patterns[:-1], patterns[1:]):
        out_walls |= blend_wall_single(first, second, interval=interval, with_symmetry=with_symmetry)
    return out_walls

def generate_symmetry(source: dict[float, "numpy array (n, 5)"], operations: list[Literal["mirror_x", "mirror_y"]|int], interval: float, pivot_3d: "numpy_array (3)" = np.zeros((3,))) -> dict[float, "numpy array (n, 5)"]:
    out = source.copy()
    offset = np.array([0.0,0.0,interval,0.0,0.0])
    counter = 1
    for o in operations:
        new_out = out.copy()
        if o in ("mirror_x", "mirror_y"):
            scale = np.array([-1.0,1.0,1.0]) if o=="mirror_x" else np.array([1.0,-1.0,1.0])
            for _, v in sorted(out.items()):
                v = movement.scale(v, scale, pivot=pivot_3d) + offset*counter
                new_out[v[0,2]] = v
            counter += 1
        elif isinstance(o, int):
            for _, v in sorted(out.items()):
                ang = 360 / o
                for r in range(1,abs(o)):
                    vrot = movement.rotate(v, ang*r, pivot=pivot_3d) + offset*counter*r
                    new_out[vrot[0,2]] = vrot
            counter += (abs(o)-1)
        else:
            raise ValueError(f"Unknown symmetry operation {o!r}")
        out = new_out
    return out

def change_wall_type(walls: "numpy array (n, 5)", new_type: str|int, direction: int = 1) -> "numpy array (n, 5)":
    if isinstance(new_type, str):
        new_type = WALL_TYPES[new_type][0]
    out = walls.copy()
    out[..., 3] = new_type
    if new_type == WALL_TYPES["crouch"][0]:
        out[..., 4] = 0
    return out



from base64 import b64decode, b64encode
from dataclasses import dataclass, field
from io import BytesIO
from time import time
from typing import Literal

from fastapi.responses import Response
from nicegui import ui, app, events
from nicegui.elements.scene_objects import Extrusion, Texture
import numpy as np
import pyperclip
import requests

from .map_render import MapScene, SettingsPanel
from .utils import GUITab, SMHInput, PrettyError, ParseInputError, PreventDefaultKeyboard, handle_errors, info, error, safe_clipboard_data
from ..utils import parse_number, pretty_time_delta, pretty_fraction, pretty_list
from .. import synth_format, movement, pattern_generation

TURBOSTACK_COUNTS = (3, 4, 5, 6, 8, 10, 12, 16, 20, 24, 32)

def make_input(label: str, value: str|float, storage_id: str, **kwargs) -> SMHInput:
    default_kwargs: dict[str, str|int] = {"tab_id": "wall_art", "width": 16}
    return SMHInput(storage_id=storage_id, label=label, default_value=value, **(default_kwargs|kwargs))

class LargeSwitch(ui.switch):
    def __init__(self, value: bool|None, storage_id: str, tooltip: str|None=None, color: str="primary", icon_unchecked: str|None=None, icon_checked: str|None=None, **kwargs):
        # ui.switch doesn't annotate tristate
        super().__init__(value=value, **kwargs)  # type: ignore
        self.bind_value(app.storage.user, f"wall_art_{storage_id}")
        self.classes("my-auto")
        self.props(f'dense size="xl" color="{color}" keep-color' + (f' unchecked-icon="{icon_unchecked}"' if icon_unchecked is not None else '') + (f' checked-icon="{icon_checked}"' if icon_checked is not None else ''))
        if tooltip is not None:
            self.tooltip(tooltip)

@ui.page("/wall_art_ref_image")
def image_proxy() -> Response:
    return Response(content=b64decode(app.storage.user.get("wall_art_ref_image", "")))

def _wall_art_tab() -> None:
    preview_scene: MapScene|None = None
    refimg_obj: Texture|None = None
    walls: synth_format.WALLS = {}
    is_dragging: bool = False

    @dataclass
    class Undo:
        undo_stack: list[tuple[str, float, synth_format.WALLS, set[float]]] = field(default_factory=list)
        redo_stack: list[tuple[str, float, synth_format.WALLS, set[float]]] = field(default_factory=list)
        max_steps: int = 50

        def reset(self):
            self.undo_stack.clear()
            self.redo_stack.clear()

        def push_undo(self, label: str):
            self.undo_stack.append((label, time(), walls.copy(), selection.sources.copy()))
            if len(self.undo_stack) > self.max_steps:
                del self.undo_stack[0]
            self.redo_stack.clear()
        
        def undo(self):
            nonlocal walls
            if not self.undo_stack:
                ui.notify("Nothing to undo", type="info", timeout=1000)
                return
            label, timestamp, undo_walls, undo_selection = self.undo_stack.pop()
            self.redo_stack.append((label, timestamp, walls.copy(), selection.sources.copy()))
            selection.clear()
            walls.clear()
            walls |= undo_walls
            selection.select(undo_selection, mode="set")
            _soft_refresh()
            ui.notify(f"Undo: {label} ({pretty_time_delta(time() - timestamp)} ago)", type="info", timeout=1000)

        def redo(self):
            nonlocal walls
            if not self.redo_stack:
                ui.notify("Nothing to redo", type="info", timeout=1000)
                return
            label, timestamp, redo_walls, redo_selection = self.redo_stack.pop()
            self.undo_stack.append((label, timestamp, walls.copy(), selection.sources.copy()))
            selection.clear()
            walls.clear()
            walls |= redo_walls
            selection.select(redo_selection, mode="set")
            _soft_refresh()
            ui.notify(f"Redo: {label} ({pretty_time_delta(time() - timestamp)} ago)", type="info", timeout=1000)

    undo = Undo()

    @dataclass
    class Selection:
        sources: set[float] = field(default_factory=set)
        cursors: dict[float, Extrusion] = field(default_factory=dict)
        drag_time: float|None = None
        offset: "np.array (3)" = field(default_factory=lambda: np.array([0.0, 0.0, 0.0]))
        mirrored: bool = False
        rotation: float = 0.0
        copy_button: bool = False
        axis_button: bool = False

        def clear(self) -> None:
            self.sources = set()
            if self.drag_time is not None:
                # everything is parented to this one
                self.cursors[self.drag_time].delete()
            else:
                for c in self.cursors.values():
                    c.delete()
            self.cursors = {}
            self.drag_time = None
            self.offset = np.array([0.0, 0.0, 0.0])
            self.mirrored = False
            self.rotation = 0

        def copy_to_clipboard(self) -> None:
            # copy everything when nothing is selected
            wall_data = walls if not self.sources else {t: walls[t] for t in self.sources}
            synth_format.export_clipboard(synth_format.ClipboardDataContainer(walls=wall_data), False)
            info(f"Copied {len(wall_data)} walls to clipboard")

        def delete(self):
            if not self.sources:
                return
            undo.push_undo("delete selection")
            ts = self.sources
            self.clear()
            for t in ts:
                del walls[t]
            _soft_refresh()
            ui.notify(f"Deleted {len(ts)} walls", type="info")

        def select(self, new_sources: set[float], mode: Literal["toggle", "expand", "set"]) -> None:
            if mode == "toggle":
                new_sources ^= self.sources
            elif mode == "expand" and self.sources:  # "expand" and nothing selected: "set"
                if len(new_sources) != 1:
                    raise RuntimeError("Cannot expand to more than one new time")
                new_t = min(new_sources)
                low = min(self.sources)
                high = max(self.sources)
                if new_t < low:
                    new_sources = self.sources | {t for t in walls if new_t <= t <= low}
                elif new_t > high:
                    new_sources = self.sources | {t for t in walls if high <= t <= new_t}
                else:
                    new_sources = self.sources | {t for t in walls if low <= t <= high}
            # else: leave new_sources as is
            self.clear()
            self.sources = new_sources
            self._update_cursors()

        def _update_cursors(self) -> None:
            if not self.sources or preview_scene is None:
                return
            first = min(self.sources)
            copy_mode = self.copy_button ^ drag_copy.value
            copy_offset = [0.0,0.0,_find_free_slot(first+self.offset[2])-first-self.offset[2],0.0,0.0] if copy_mode else 0.0
            if self.drag_time is None:
                # re-create cursors
                for c in self.cursors.values():
                    try:
                        c.delete()
                    except KeyError:
                        # scene sometimes loses track of objects, so use internal delete
                        c._delete()
                preview_settings = sp.parse_settings()
                with preview_scene:
                    pivot_3d = walls[first][0,:3]
                    scale_3d = np.array([1.0, -1.0 if self.mirrored else 1.0, 1.0])
                    for t in self.sources:
                        w = walls[t] + copy_offset
                        w = movement.rotate(w, angle=self.rotation, pivot=pivot_3d)
                        w = movement.scale(w, scale_3d=scale_3d, pivot=pivot_3d)
                        w = movement.offset(w, self.offset)
                        e = preview_scene.wall_extrusion(w, preview_settings.wall.size * time_scale.parsed_value).draggable()
                        if copy_mode:
                            e.material(copy_color.value, copy_opacity.parsed_value)
                        else:
                            e.material(move_color.value, move_opacity.parsed_value)
                        self.cursors[t] = e
                preview_scene.props('drag_constraints=""')
            else:
                w = walls[self.drag_time][0] + copy_offset
                scene_pos = preview_scene.to_scene(w[:3] + self.offset)
                # move cursor
                cur = self.cursors[self.drag_time]
                cur.move(*scene_pos)
                rot = w[4]+self.rotation
                if self.mirrored:
                    cur.scale(1,-1,1)
                    cur.rotate(np.deg2rad(90), np.deg2rad(180 + rot), 0)
                else:
                    cur.scale(1,1,1)
                    cur.rotate(np.deg2rad(90), np.deg2rad(180 - rot), 0)
                # update drag constraints
                if self.axis_button ^ axis_z.value:
                    scene_time_step = time_step.parsed_value*time_scale.parsed_value
                    preview_scene.props(f'drag_constraints="x={scene_pos[0]},z={scene_pos[2]},y=Math.round(y/({scene_time_step}))*({scene_time_step})"')
                else:
                    preview_scene.props(f'drag_constraints="y={scene_pos[1]}"')
            
            for c in self.cursors.values():
                if copy_mode:
                    c.material(copy_color.value, copy_opacity.parsed_value)
                else:
                    c.material(move_color.value, move_opacity.parsed_value)

        def move(self, offset: "numpy array (3)") -> None:
            self.offset += offset
            self._update_cursors()

        def rotate(self, rotation: float) -> None:
            self.rotation += rotation if not self.mirrored else -rotation
            self._update_cursors()

        def mirror(self, horizontal: bool) -> None:
            self.mirrored = not self.mirrored
            if horizontal:
                self.rotation += 180
            self._update_cursors()

        def set_copy_button(self, copy_button: bool) -> None:
            if copy_button != self.copy_button:
                self.copy_button = copy_button
                self._update_cursors()

        def set_axis_button(self, axis_button: bool) -> None:
            if axis_button != self.axis_button:
                self.axis_button = axis_button
                self._update_cursors()

        @handle_errors
        def apply(self) -> None:
            nonlocal walls
            copy_mode = self.copy_button ^ drag_copy.value
            ops = [l for v, l in [(copy_mode, "copy"), (self.rotation, "rotate"), (self.mirrored, "mirror"), (np.any(self.offset), "offset")] if v]
            undo.push_undo(f"{pretty_list(ops)} {len(self.sources)} walls")
            first = min(self.sources)
            copy_offset = [0.0,0.0,_find_free_slot(first+self.offset[2])-first-self.offset[2],0.0,0.0] if copy_mode else 0.0
            pivot_3d = walls[self.drag_time if self.drag_time is not None else first][0,:3]
            scale_3d = np.array([1.0, -1.0 if self.mirrored else 1.0, 1.0])
            new_walls = {}
            for t in sorted(self.sources):
                w = walls[t]+copy_offset if copy_mode else walls.pop(t)
                w = movement.rotate(w, angle=self.rotation, pivot=pivot_3d)
                w = movement.scale(w, scale_3d=scale_3d, pivot=pivot_3d)
                w = movement.offset(w, self.offset)
                new_walls[w[0,2]] = w
            if turbostack.value and copy_mode:
                extra_count = TURBOSTACK_COUNTS[turbostack.value-1] - 2  # subtract base and copy
                stack_walls = new_walls.copy()
                for _ in range(extra_count):
                    stack_new = {}
                    for t in sorted(stack_walls):
                        w = stack_walls[t]+copy_offset
                        w = movement.rotate(w, angle=self.rotation, pivot=pivot_3d)
                        w = movement.scale(w, scale_3d=scale_3d, pivot=pivot_3d)
                        w = movement.offset(w, self.offset)
                        stack_new[w[0,2]] = w
                    new_walls |= stack_new
                    stack_walls = stack_new
            new_sources = set(new_walls)

            sym_ops: list[Literal["mirror_x", "mirror_y"]|int] = []
            if mirror_x.value:
                sym_ops.append("mirror_x")
            if mirror_y.value:
                sym_ops.append("mirror_y")
            if rotsym_direction.value is not None:
                rsym = rotsym.value * (-1 if rotsym_direction.value else 1)
                if rotate_first.value:
                    sym_ops.insert(0, int(rsym))
                else:
                    sym_ops.append(int(rsym))
            sym_interval = symmetry_step.parsed_value
            new_walls |= pattern_generation.generate_symmetry(source=new_walls, operations=sym_ops, interval=sym_interval)
                
            for _, w in sorted(new_walls.items()):
                _insert_wall(w, displace_forward=self.offset[2]>0)
            _soft_refresh()
            self.select(new_sources, "set")

        @handle_errors
        def start_drag(self, object_id: str) -> None:
            if preview_scene is None:
                return
            copy_mode = self.copy_button ^ drag_copy.value
            for t, c in self.cursors.items():
                if c.id == object_id:
                    self.drag_time = t
                    break
            else:
                return
            for t, c in self.cursors.items():
                if t != self.drag_time:
                    c.delete()
            pivot = walls[self.drag_time]
            preview_settings = sp.parse_settings()
            with self.cursors[self.drag_time]:
                for t in self.sources:
                    if t != self.drag_time:
                        w = walls[t]
                        relative = np.array([[0.0,0.0,0.0, w[0,3], 0.0]])
                        e = preview_scene.wall_extrusion(relative, preview_settings.wall.size * time_scale.parsed_value)
                        offset = movement.rotate(w-pivot, 180-pivot[0,4])
                        e.move(offset[0,0], offset[0,1], -(w[0,2]-pivot[0,2])*time_scale.parsed_value).rotate(0,0,np.deg2rad(w[0,4]-pivot[0,4]))
                        if copy_mode:
                            e.material(copy_color.value, copy_opacity.parsed_value/2)
                        else:
                            e.material(move_color.value, move_opacity.parsed_value/2)
                        self.cursors[t] = e
            self._update_cursors()

        def end_drag(self, xyt: tuple[float, float, float]) -> None:
            if self.drag_time not in walls:
                self.clear()
                return
            self.offset = np.array(xyt) - walls[self.drag_time][0,:3]
            self.apply()

    selection = Selection()
    last_ctrl_press = None
    last_shift_press = None
    @handle_errors
    def _on_key(e: events.KeyEventArguments) -> None:
        nonlocal last_ctrl_press, last_shift_press
        if e.key.control and not e.action.repeat:
            if e.action.keydown:
                # on CTRL double-tap, toggle drag_copy value
                if last_ctrl_press is not None and time() - last_ctrl_press < 0.5:
                    drag_copy.set_value(not drag_copy.value)
                    last_ctrl_press = None
                    return
                else:
                    last_ctrl_press = time()
            selection.set_copy_button(e.action.keydown)
        else:
            # if any other key is pressed/release clear double-tap
            last_ctrl_press = None

        if e.key.shift and not e.action.repeat:
            if e.action.keydown:
                # on SHIFT double-tap, toggle axis_z value
                if last_shift_press is not None and time() - last_shift_press < 0.5:
                    axis_z.set_value(not axis_z.value)
                    last_shift_press = None
                    return
                else:
                    last_shift_press = time()
            selection.set_axis_button(e.action.keydown)
        else:
            # if any other key is pressed/release clear double-tap
            last_shift_press = None

        if not e.action.keydown:
            return
        try:
            # note: don't use key.code, as that doesn't account for keyboard layout
            key_name = e.key.name.upper()  # key.name is upper/lowercase depending on shift
            # CTRL-independent
            if e.key.number in range(1, len(synth_format.WALL_LOOKUP)+1):
                assert e.key.number is not None  # mypy doesn't unstand the range above
                wall_type = sorted(synth_format.WALL_LOOKUP)[e.key.number-1]
                _spawn_wall(wall_type=wall_type, change_selection=e.modifiers.ctrl, extend_selection=e.modifiers.shift)
            elif e.key.page_up or e.key.page_down:
                selection.move(np.array([0.0,0.0,(e.key.page_up-e.key.page_down)*time_step.parsed_value])*offset_step.parsed_value)
            elif e.key.enter or key_name == " ":
                selection.apply()
            # CTRL: Yes
            elif e.modifiers.ctrl:
                if key_name == "A":
                    # select all
                    selection.select(set(walls), mode="set")
                    # clear text selection
                    ui.run_javascript("""
                    var sel = window.getSelection ? window.getSelection() : document.selection;
                        if (sel) {
                            if (sel.removeAllRanges) {
                                sel.removeAllRanges();
                            } else if (sel.empty) {
                                sel.empty();
                            }
                        }
                    """)
                elif key_name == "C":
                    selection.copy_to_clipboard()
                elif key_name == "X":
                    selection.copy_to_clipboard()
                    if selection.sources:
                        selection.delete()
                    else:
                        undo.push_undo("cut to clipboard")
                        walls.clear()
                        _soft_refresh()
                elif key_name == "V":
                    _paste()
                elif key_name == "Z":
                    undo.undo()
                elif key_name == "Y":
                    undo.redo()
            # CTRL: No
            elif key_name == "C":
                _reset_camera()
            elif e.key.escape:
                selection.select(set(), "set")
            elif key_name == "R":
                _compress()
            elif key_name == "B":
                _open_blend_dialog()
            elif key_name == "E":
                ordered_keys = sorted(walls)
                if not ordered_keys:
                    return
                if not selection.sources:
                    selection.select({ordered_keys[0]}, mode="set")
                elif max(selection.sources) in ordered_keys[:-1]:
                    selection.select({ordered_keys[ordered_keys.index(max(selection.sources))+1]}, mode="set" if not e.modifiers.shift else "expand")
            elif key_name == "Q":
                ordered_keys = sorted(walls)
                if not ordered_keys:
                    return
                if not selection.sources:
                    selection.select({ordered_keys[-1]}, mode="set")
                elif min(selection.sources) in ordered_keys[1:]:
                    selection.select({ordered_keys[ordered_keys.index(min(selection.sources))-1]}, mode="set" if not e.modifiers.shift else "expand")
            elif not selection.sources:
                return
            elif e.key.delete or e.key.backspace:
                selection.delete()
            elif key_name in "AD":
                selection.rotate((angle_step.parsed_value if not e.modifiers.shift else 90.0) * (-1 if key_name == "D" else 1))
            elif key_name in "WS":
                selection.mirror(horizontal=(key_name=="S"))
            elif e.key.is_cursorkey:
                selection.move(np.array([(e.key.arrow_right-e.key.arrow_left),(e.key.arrow_up-e.key.arrow_down),0.0])*offset_step.parsed_value)
        except ParseInputError as pie:
            error(f"Error parsing setting: {pie.input_id}", pie, data=pie.value)
            return
    PreventDefaultKeyboard(on_key=_on_key).bind_active_from(app.storage.user, "active_tab", backward=lambda v: v=="wall_art")
    with ui.card():
        with ui.row():
            with ui.row():
                axis_z = LargeSwitch(False, "axis", "(Double-tab SHIFT) Change movement axis between X/Y and Time (inverts with SHIFT held)", color="info", icon_unchecked="open_with", icon_checked="schedule")
                drag_copy = LargeSwitch(False, "drag_copy", "(Double-tap CTRL) Switch default drag behavior between move and copy (inverts with CTRL held).", color="positive", icon_unchecked="start", icon_checked="call_split", on_change=selection._update_cursors)
                displace = LargeSwitch(False, "displace", "Displace existing walls when moving in time instead of replacing them.", color="warning", icon_unchecked="cancel", icon_checked="move_up")
                time_step = make_input("Time Step", "1/64", "time_step", suffix="b", tooltip="Time step for adding walls or moving via dragg or (page-up)/(page-down)")
                offset_step = make_input("Offset Step", "1", "offset_step", suffix="sq", tooltip="Step for moving via (arrow keys)")
                angle_step = make_input("Angle Step", "15", "angle_step", suffix="°", tooltip="Step for rotation via (A)/(D)")
            with ui.expansion("Symmetry", icon="flip").props("dense") as sym_exp:
                with ui.row():
                    symmetry_step = make_input("Interval", "1/4", "symmetry_step", suffix="b", tooltip="Time step for symmetry copies")
                    with LargeSwitch(False, "rotate_first", color="secondary", icon_unchecked="flip", icon_checked="adjust") as rotate_first:
                        ui.tooltip().bind_text_from(rotate_first, "value", lambda rf: "Rotate first, then mirror" if rf else "Mirror first, then rotate")
                    ui.separator().props("vertical")
                    mirror_x = LargeSwitch(False, "mirror_x", "Mirror across X axis, ie left-right", color="negative", icon_unchecked="align_horizontal_left", icon_checked="align_horizontal_center")
                    mirror_y = LargeSwitch(False, "mirror_y", "Mirror across Y axis, ie up-down", color="positive", icon_unchecked="align_vertical_bottom", icon_checked="align_vertical_center")
                with ui.row():
                    ui.label("Rotation: ").classes("my-auto")
                    rotsym_direction = LargeSwitch(None, "rotsym_direction", "Rotation direction", color="white", icon_unchecked="rotate_left", icon_checked="rotate_right").props('toggle-indeterminate indeterminate-icon="cancel" icon-color="black"')
                    ui.tooltip("Number of rotational symmetry. Note that mirror in both X and Y overlaps with even symmetries, ie 2x/4x/etc")
                    rotsym = ui.slider(min=2, max=12, value=2).props('snap markers selection-color="transparent" color="secondary" track-size="2px" thumb-size="25px"').classes("w-24").bind_value(app.storage.user, "wall_art_rotsym").bind_enabled_from(rotsym_direction, "value", backward=lambda v: v is not None)
                    ui.label().classes("my-auto w-8").bind_text_from(rotsym, "value", backward=lambda v: f"x{v}").bind_visibility_from(rotsym_direction, "value", backward=lambda v: v is not None)
                with ui.row():
                    ui.label("TurboStack: ").classes("my-auto")
                    ui.tooltip("Create stacked copies")
                    turbostack = ui.slider(min=0, max=len(TURBOSTACK_COUNTS), value=0).props('snap markers selection-color="transparent" color="lime" track-size="2px" thumb-size="25px"').classes("w-24").bind_value(app.storage.user, "wall_art_turbostack")
                    ui.label().classes("my-auto w-8").bind_text_from(turbostack, "value", backward=lambda v: f"{TURBOSTACK_COUNTS[v-1]}x" if v else "off")#.bind_visibility_from(rotsym_direction, "value", backward=lambda v: v is not None)
                def _update_symex(_) -> str:
                    enabled = []
                    if mirror_x.value:
                        enabled.append("X")
                    if mirror_y.value:
                        enabled.append("Y")
                    if rotsym_direction.value is not None:
                        rotdir = "↻" if rotsym_direction.value else "↺"
                        if rotate_first.value:
                            enabled.insert(0, f"{rotdir}{rotsym.value}")
                        else:
                            enabled.append(f"{rotdir}{rotsym.value}")
                    stack_str = f"Stack: {TURBOSTACK_COUNTS[turbostack.value-1]}x, " if turbostack.value else ""
                    return f"{stack_str}Symmetry: {', '.join(enabled) or 'off'}"
                for inp in (mirror_x, mirror_y, rotsym_direction, rotate_first, turbostack):
                    inp.bind_value_to(sym_exp, "text", forward=_update_symex)
            with ui.expansion("Reference", icon="image").props("dense"):
                ui.tooltip("Display a reference image to align wall art.")
                refimg_url = ui.input("Reference Image URL").props("dense").classes("w-full").bind_value(app.storage.user, "wall_art_ref_image_url").tooltip("Direct URL to image file. Press the download button below to download.")
                with ui.row():
                    def _clear_image() -> None:
                        app.storage.user["wall_art_ref_image"] = ""
                        ui.notify("Reference image cleared. Click APPLY to apply.", type="positive")
                    ui.button(icon="clear", on_click=_clear_image, color="negative").props("outline").tooltip("Clear image data").classes("w-10")
                    @handle_errors
                    def _download_image() -> None:
                        url = refimg_url.value
                        if not url:
                            raise PrettyError(msg="Set an URL above")
                        else:
                            try:
                                r = requests.get(url)
                                r.raise_for_status()
                            except requests.RequestException as req_exc:
                                raise PrettyError(msg="Downloading image failed", exc=req_exc, data=url) from req_exc
                            data = r.content
                            app.storage.user["wall_art_ref_image"] = b64encode(data).decode()
                            ui.notify(f"Downloaded reference image ({len(data)/1024:.1f} KiB). Click APPLY to apply.", type="positive")
                    ui.button(icon="cloud_download", on_click=_download_image, color="positive").props("outline").tooltip("Download image from URL").classes("w-10")
                    def _upload_image(e: events.UploadEventArguments) -> None:
                        upl: ui.upload = e.sender  # type:ignore
                        upl.reset()
                        data = e.content.read()
                        app.storage.user["wall_art_ref_image"] = b64encode(data).decode()
                        ui.notify(f"Uploaded reference image ({len(data)/1024:.1f} KiB). Click APPLY to apply.", type="positive")
                    refimg_upload = ui.upload(label="Upload File", multiple=True, auto_upload=True, on_upload=_upload_image).props('outline color="positive" accept="image/*"').classes("w-28")
                    with refimg_upload.add_slot("list"):
                        pass
                with ui.row():
                    refimg_width = make_input("Width", "16", "ref_width", suffix="sq", tooltip="Width of the reference image in sq")
                    refimg_height = make_input("Height", "12", "height", suffix="sq", tooltip="Height of the reference image in sq")
                    refimg_opacity = make_input("Opacity", "0.5", "ref_opacity", tooltip="Opacity of the image (0-1). 1=completely opaque, 0=completely transparent")
                with ui.row():
                    refimg_x = make_input("X", "0", "ref_x", suffix="sq", tooltip="Center X of the reference image in sq")
                    refimg_y = make_input("Y", "0", "ref_y", suffix="sq", tooltip="Center Y of the reference image in sq")
                    refimg_t = make_input("Time", "1/4", "ref_time", suffix="b", tooltip="Time of the reference image in beats")
                refimg_apply_button = ui.button("Apply").props("outline")
            with ui.expansion("Preview setttings", icon="palette").props("dense"):
                sp = SettingsPanel()
                ui.separator()
                with ui.row():
                    ui.icon("highlight_alt", size="3em").tooltip("Appearance of selected walls")
                    move_color = ui.color_input("Move", value="#888888", preview=True).props("dense").classes("w-28").bind_value(app.storage.user, "wall_art_move_color")
                    move_color.button.style("color: black")
                    move_opacity = make_input("Opacity", "0.5", "move_opacity")
                    copy_color = ui.color_input("Copy", value="#00ff00", preview=True).props("dense").classes("w-28").bind_value(app.storage.user, "wall_art_copy_color")
                    copy_color.button.style("color: black")
                    copy_opacity = make_input("Opacity", "0.5", "copy_opacity")
                ui.separator()
                with ui.row():
                    ui.icon("preview", size="3em").tooltip("Change size and scaling of preview")
                    scene_width = make_input("Width", "800", "width", tab_id="preview", suffix="px", tooltip="Width of the preview in px")
                    scene_height = make_input("Height", "600", "height", tab_id="preview", suffix="px", tooltip="Height of the preview in px")
                    time_scale = make_input("Time Scale", "64", "time_scale", tab_id="preview", tooltip="Ratio between XY and time")
                    frame_length = make_input("Frame Length", "2", "frame_length", tab_id="preview", suffix="b", tooltip="Number of beats to draw frames for")
                ui.separator()
                with ui.row():
                    ui.icon("camera_indoor", size="3em").tooltip("Change home position of camera")
                    cam_height = make_input("Cam Height", "2", "cam_height", tab_id="preview", suffix="sq", tooltip="Default camera height")
                    cam_time = make_input("Cam Time", "1/2", "cam_time", tab_id="preview", suffix="b", tooltip="Default camera time center")
                    cam_distance = make_input("Cam Distance", "1", "cam_distance", tab_id="preview", suffix="b", tooltip="Default camera distance from center")
                preview_apply_button = ui.button("Apply").props("outline")
        def _find_free_slot(t: float) -> float:
            while t in walls:
                t = np.round(t/time_step.parsed_value + 1)*time_step.parsed_value
            return t

        @handle_errors
        def _insert_wall(w: "np.array (1,5)", displace_forward: bool = False) -> None:
            if not displace.value:
                walls[w[0,2]] = w
                return
            pending = w
            displace_dir = -1 if displace_forward else 1
            while pending[0,2] in walls:
                walls[pending[0,2]], pending = pending, walls[pending[0,2]]
                pending[0,2] = np.round(pending[0,2]/time_step.parsed_value + displace_dir)*time_step.parsed_value
            walls[pending[0,2]] = pending

        @handle_errors
        def _spawn_wall(wall_type: int, change_selection: bool = False, extend_selection: bool = False) -> None:
            if change_selection:
                if not selection.sources:
                    return
                undo.push_undo(f"change {len(selection.sources)} walls to {synth_format.WALL_LOOKUP[wall_type]}")
                for s in selection.sources:
                    walls[s] = pattern_generation.change_wall_type(walls[s], wall_type)
                _soft_refresh()
                selection.select(set(), "toggle")
            else:
                undo.push_undo(f"add {synth_format.WALL_LOOKUP[wall_type]}")
                new_t = 0.0 if not selection.sources else max(selection.sources) + time_step.parsed_value
                if not displace.value:
                    new_t = _find_free_slot(max(selection.sources, default=0.0))
                _insert_wall(np.array([[0.0,0.0,new_t,wall_type,0.0]]))
                _soft_refresh()
                selection.select({new_t}, mode="set" if not extend_selection else "toggle")

        @handle_errors
        def _soft_refresh() -> None:
            nonlocal refimg_obj
            preview_settings = sp.parse_settings()
            if preview_scene is None:
                draw_preview_scene.refresh()
                _reset_camera()
            if refimg_obj is not None:
                refimg_obj.delete()
                refimg_obj = None
            if preview_scene is not None:
                if refimg_url.value:
                    with preview_scene:
                        coords = np.array([[[-1/2,0,1/2],[1/2,0,1/2]],[[-1/2,0,-1/2],[1/2,0,-1/2]]]) * [refimg_width.parsed_value,0,refimg_height.parsed_value]
                        pos = (refimg_x.parsed_value, refimg_t.parsed_value*time_scale.parsed_value, refimg_y.parsed_value)
                        opacity = refimg_opacity.parsed_value
                        if app.storage.user.get("wall_art_ref_image"):
                            # add parameter to bypass cache
                            refimg_obj = preview_scene.texture(f"/wall_art_ref_image?nocache={time()}", coords.tolist()).move(*pos).material(opacity=opacity)
                wall_data = synth_format.DataContainer(walls=walls)
                preview_scene.render(wall_data, preview_settings)

        @handle_errors
        def _reset_camera() -> None:
            if preview_scene is None:
                return
            preview_scene.move_camera(
                0, (cam_time.parsed_value-cam_distance.parsed_value)*time_scale.parsed_value, cam_height.parsed_value,
                0, cam_time.parsed_value*time_scale.parsed_value, cam_height.parsed_value,
                duration=0,
            )

        @handle_errors
        def _on_click(e: events.SceneClickEventArguments) -> None:
            if preview_scene is None:
                return
            nonlocal is_dragging
            if is_dragging or e.alt:
                is_dragging = False
                return
            try:
                for h in e.hits:
                    if h.object_id in preview_scene.wall_lookup:
                        t = preview_scene.wall_lookup[h.object_id]
                        selection.select({t}, mode="toggle" if e.ctrl else "expand" if e.shift else "set")
                        return
                selection.select(set(), "set")
            except ParseInputError as pie:
                error(f"Error parsing preview setting: {pie.input_id}", pie, data=pie.value)
                return
        def _on_dstart(e: events.SceneDragEventArguments):
            nonlocal is_dragging
            is_dragging = True
            selection.start_drag(e.object_id)

        def _on_dend(e: events.SceneDragEventArguments):
            if is_dragging:
                selection.end_drag((e.x, e.z, round((e.y / time_scale.parsed_value)/time_step.parsed_value)*time_step.parsed_value))

        @ui.refreshable
        @handle_errors
        def draw_preview_scene() -> None:
            nonlocal preview_scene
            w = int(scene_width.parsed_value)
            h = int(scene_height.parsed_value)
            l = int(frame_length.parsed_value)
            t = time_scale.parsed_value
            preview_scene = MapScene(width=w, height=h, frame_length=l, time_scale=t, on_click=_on_click, on_drag_start=_on_dstart, on_drag_end=_on_dend)
            _reset_camera()
            _soft_refresh()

        with ui.row():
            with ui.column():
                with ui.button(icon="undo", color="negative", on_click=undo.undo).props("outline").style("width: 36px").bind_enabled_from(undo, "undo_stack", backward=bool):
                    ui.tooltip().bind_text_from(undo, "undo_stack", backward=lambda us: f"Undo '{us[-1][0]}' (CTRL+Z) [{len(us)} steps]" if us else "Undo (CTRL+Z)")
                with ui.button(icon="redo", color="positive", on_click=undo.redo).props("outline").style("width: 36px").bind_enabled_from(undo, "redo_stack", backward=bool):
                    ui.tooltip().bind_text_from(undo, "redo_stack", backward=lambda rs: f"Redo '{rs[-1][0]}' (CTRL+Y) [{len(rs)} steps]" if rs else "Redo (CTRL+Y)")
                ui.separator()
                @handle_errors
                def _paste() -> None:
                    with safe_clipboard_data(use_original=False, write=False) as data:
                        undo.push_undo("paste from clipboad")
                        walls.update(data.walls)
                    selection.select(set(), mode="toggle")
                    _soft_refresh()
                    info(f"Added {len(data.walls)} walls from clipboard")
                with ui.button(icon="content_paste", color="positive", on_click=_paste).style("width: 36px"):
                    ui.tooltip("Add walls from clipboard (CTRL+V)")
                with ui.button(icon="deselect", on_click=lambda: selection.select(set(), "set")).props("outline").style("width: 36px").bind_enabled_from(selection, "sources", backward=bool):
                    ui.tooltip("Deselect (ESC)")
                with ui.button(icon="select_all", on_click=lambda: selection.select(set(walls), "set")).props("outline").style("width: 36px"):
                    ui.tooltip("Select all (CTRL+A)")
                with ui.button(icon="content_copy", on_click=selection.copy_to_clipboard).style("width: 36px"):
                    ui.tooltip("Copy (CTRL+C)")
                def _clear() -> None:
                    undo.reset()
                    selection.clear()
                    walls.clear()
                    _soft_refresh()
                with ui.button(icon="clear", color="negative", on_click=_clear).props("outline").style("width: 36px"):
                    ui.tooltip("Clear everything (includes undo steps)")
                ui.separator()
                @handle_errors
                def _compress() -> None:
                    undo.push_undo("compress")
                    new_sources = set()
                    for i, (t, w) in enumerate(sorted(walls.items())):
                        w = w.copy()
                        del walls[t]
                        w[...,2] = i * time_step.parsed_value
                        walls[w[0,2]] = w
                        if t in selection.sources:
                            new_sources.add(w[0,2])
                    selection.select(new_sources, mode="set")
                    _soft_refresh()

                with ui.button(icon="compress", on_click=_compress).props("outline").style("width: 36px"):
                    ui.tooltip("Compress wall spacing to time step (effects all walls)")
                with ui.dialog() as blend_dialog, ui.card():
                    ui.label("Blend between patterns")
                    blend_pattern = ui.select({}, label="Choose a detected pattern:")
                    # >1 option: show selector
                    blend_pattern.bind_visibility_from(blend_pattern, "options", backward=lambda opts: len(opts)!=1)
                    # 1 option: show just text
                    ui.label("").bind_visibility_from(blend_pattern, "options", backward=lambda opts: len(opts)==1).bind_text_from(
                        blend_pattern, "options", backward=lambda opts: "Detected: " + next(iter(opts.values())) if opts else ""
                    )
                    with ui.row():
                        blend_interval = make_input("Interval", "1/2", "blend_interval", tooltip="Interval between blending steps", suffix="b")
                        def _do_blend() -> None:
                            nonlocal walls
                            wc, pc, pl = blend_pattern.value
                            error_data = [w.tolist() for w in walls.values()]
                            if len(walls) != wc*pc:
                                error(f"Wall count changed! Expected {pc*wc}, found {len(walls)}", data=error_data)
                                return
                            try:
                                patterns = np.array([w[0] for _, w in sorted(walls.items())]).reshape((pc, wc, 5))
                                interval = blend_interval.parsed_value
                            except ValueError as ve:
                                error(f"Could not split up {len(walls)} walls into {pc} patterns with {wc} wall{'s'*(wc!=1)} each", ve, data=error_data)
                                return
                            pattern_deltas = np.diff(patterns[:,0,2])
                            if pattern_deltas.max() <= interval:
                                error(
                                    f"Blend interval ({pretty_fraction(interval)}b) must be greater than largest pattern distance ({pretty_fraction(pattern_deltas.max())}b), else blending will do nothing.",
                                    data=error_data,
                                )
                                return

                            if interval <= pl:
                                ui.notify(
                                    f"Blend interval ({pretty_fraction(interval)}b) is shorter than pattern length ({pretty_fraction(pl)}b), leading to overlaps. This may result in strange results.",
                                    type="warning",
                                )

                            small_deltas = np.sum(pattern_deltas < interval)  # equal is a "regular" usecase, ie when re-blending some parts
                            if small_deltas: 
                                ui.notify(
                                    f"Blend interval ({pretty_fraction(interval)}b) is smaller than {small_deltas} of {patterns.shape[0]-1} pattern distances. Blending will do nothing between these.",
                                    type="warning"
                                )
                            undo.push_undo("blend")
                            try:
                                walls |= pattern_generation.blend_walls_multiple(patterns, interval=interval)
                            except ValueError as ve:
                                error("Error blending walls", ve, data=error_data)
                                return
                            added = len(walls)//patterns.shape[1] - patterns.shape[0]
                            ui.notify(f"Created {added} additional pattern{'s'*(added!=1)} between the existing {patterns.shape[0]}", type="info")
                            _soft_refresh()
                            blend_dialog.close()
                        ui.button(icon="blender", on_click=_do_blend).props("outline").classes("w-16 h-10")

                def _open_blend_dialog() -> None:
                    error_data = [w.tolist() for w in walls.values()]
                    if len(walls) < 2:
                        error("Need at least two walls to do blending", data=error_data)
                        return
                    # autodetect patterns
                    try:
                        detected_patterns = pattern_generation.find_wall_patterns(walls)
                    except ValueError as ve:
                        error(f"Could not detect patterns in walls: {ve}", data=error_data)
                        return
                    blend_dialog.open()
                    blend_pattern.set_options({
                        (w, p, l): (
                            f"{p} patterns, each with {w} walls and {pretty_fraction(l)}b long"
                            if w != 1 else f"{p} instances of a single wall"
                        )
                        for w, p, l in detected_patterns
                    })
                    blend_pattern.set_value(detected_patterns[0])
                    blend_interval.update()  # workaround for nicegui#2149

                with ui.button(icon="blender", on_click=_open_blend_dialog, color="info").style("width: 36px"):
                    ui.tooltip("Blend wall patterns (effects all walls)")
            draw_preview_scene()
            with ui.column():
                with ui.button(icon="camera_indoor", on_click=_reset_camera).style("width: 36px"):
                    ui.tooltip("Reset camera position (C), can be adjusted in preview settings")
                ui.separator()
                with ui.button(icon="delete", color="negative", on_click=selection.delete).style("width: 36px").bind_enabled_from(selection, "sources", backward=bool):
                    ui.tooltip("Delete selection (Delete/Backspace)")
                for k, (i, t) in enumerate(sorted(synth_format.WALL_LOOKUP.items())):
                    def _add_wall(e: events.GenericEventArguments, wall_type=i) -> None:
                        # python closure doesn't work as needed here, so enclose i as default param instead
                        _spawn_wall(wall_type=wall_type, change_selection=e.args["ctrlKey"], extend_selection=e.args["shiftKey"])
                    with ui.button(color="positive").on("click", _add_wall).classes("p-2"):
                        ui.tooltip(f"({k+1}) Spawn '{t}' wall (after selection), hold CTRL to change wall type instead")
                        v = synth_format.WALL_VERTS[t]
                        # draw wall vertices as svg
                        content = f'''
                            <svg viewBox="-10 -10 20 20" width="20" height="20" xmlns="http://www.w3.org/2000/svg">
                                <polygon points="{' '.join(f"{-x},{y}" for x,y in v)}" stroke="white"/>
                            </svg>
                        '''
                        ui.html(content)

        refimg_apply_button.on("click", draw_preview_scene.refresh)
        preview_apply_button.on("click", draw_preview_scene.refresh)

        with ui.dialog() as key_dialog, ui.card():
            with ui.tabs() as control_tabs:
                ui.tab("Camera", icon="cameraswitch")
                ui.tab("General Shortcuts", icon="keyboard")
                ui.tab("Edit Shortcuts", icon="tune")
            with ui.tab_panels(control_tabs, value="Camera"):
                with ui.tab_panel("Camera"):
                    ui.markdown("""
                        Use left mouse to rotate the camera, right mouse to move. Scroll wheel zooms.  
                        To turn the camera *without* deselecting, hold down ALT.  
                        🇨 resets the camera to the position configured in the settings.

                        Click on a wall to select it. Multiple walls can be selected by CTRL-Click (to add), or SHIFT-Click (expand selection to clicked wall).  
                        Then drag it around using left mouse and/or use one of the edit shortcut.  
                        Holding CTRL creates a (edited) copy in the next free slot instead of moving.  
                        Holding SHIFT drags in time instead of X/Y.
                    """)
                with ui.tab_panel("General Shortcuts"):
                    ui.markdown("""
                        |Key|Function|
                        |-|-|
                        |ESC|Deselect all|
                        |CTRL+🇦|Select all|
                        |CTRL+🇨|Copy to clipboard (selection or everything)|
                        |CTRL+🇽|Cut to clipboard (selection or everything)|
                        |CTRL+🇻|Add walls from clipboard (overrides existing)|
                        |CTRL+🇿|Undo last operation|
                        |CTRL+🇾|Redo last operation|
                        |Tap SHIFT twice|Toggle drag axis between time and X/Y| 
                        |Tap CTRL twice|Toggle drag behavior between move and copy| 
                        |🇶/🇪|Select previous/next Wall (SHIFT: Expand selection)|
                        |🇷|Compress all walls to timestep|
                        |🇧|Open Blender|
                    """)
                with ui.tab_panel("Edit Shortcuts"):
                    ui.markdown("""
                        If you use keyboard shortcuts instead of dragging, press enter, space or click the shadow to apply.

                        Note: If you hold down CTRL to copy, most of these will not work.  
                        Either let down of CTRL to use them (and then press it again), or toggle the copy mode via CTRL double-tap.  
                        You have to move the mouse around slightly to correct the drag position after switching CTRL state.

                        |Key|Function|
                        |-|-|
                        |1️⃣-8️⃣|Spawn & select wall (SHIFT: Add to current selection)|
                        |CTRL+1️⃣-8️⃣|Change wall type|
                        |`Del`/Backspace|Delete selection|
                        |🇦/🇩|Rotate by step (SHIFT: 90 degree)|
                        |🇼|Mirror on Y axis (up-down)|
                        |🇸|Mirror on X axis (left-right)|
                        |⬅️➡️⬆️⬇️|Offset X/Y|
                        |Page Up/Down|Offset Time|
                        |Enter/Space|Apply (CTRL: Copy to next free slot)|
                    """)
        with ui.button("Controls", icon="keyboard", on_click=key_dialog.open, color="info").classes("cursor-help"):
            ui.tooltip("Show controls")

wall_art_tab = GUITab(
    name="wall_art",
    label="Wall Art",
    icon="wallpaper",
    content_func=_wall_art_tab,
)


import datetime

from nicegui import app, ui
import requests
import logging

from .utils import GUITab, error
from .. import __version__
from ..utils import pretty_time_delta

releases = None

def _version_tab():
    @ui.refreshable
    def release_list():
        if releases == "error":
            ui.label("Error requesting version information")
        elif releases:
            now = datetime.datetime.now(datetime.timezone.utc)
            new = True
            for r in releases[:10]:
                try:
                    time_delta = now - datetime.datetime.fromisoformat(r["created_at"]).replace(tzinfo=datetime.timezone.utc)
                    name = r["name"]
                    v = r["tag_name"][1:]
                    url = r["html_url"]
                    body = r["body"]
                except:
                    ui.label("Error parsing version information")
                else:
                    color = "positive" if new else "auto"
                    icon = "stars" if new else "check_circle"
                    if v == __version__:
                        color = "primary"
                        icon = "play_circle"
                        new = False
                    with ui.expansion(f'{name} ({pretty_time_delta(time_delta.total_seconds())} ago)', icon=icon).props(f'group="ver_hist" header-class="bg-{color} w-auto"'):
                        with ui.row():
                            ui.label("Github link:")
                            ui.link(url, target=url, new_tab=True) 
                        with ui.card().props("bordered"):
                            ui.markdown(r["body"])
        elif releases is not None:
            with ui.row():
                ui.spinner()
                ui.label("Requesting...")
        else:
            ui.label("To protect your privacy, this does not automatically contact github.com. Press the button or enable auto-check below.")

    def check():
        global releases
        try:
            releases = []
            release_list.refresh()
            r = requests.get("https://api.github.com/repos/adosikas/synth_mapping_helper/releases")
            r.raise_for_status()
            new_data = r.json()
            if not isinstance(new_data, list):
                releases = "error"
                release_list.refresh()
                error("github.com returned unexpected data", data=new_data)
                return
            releases = new_data
            if releases[0]["tag_name"][1:] != __version__:
                ui.notify(
                    f"A new update is available: {releases[0]['name']}",
                    caption="To update, run the installer again",
                    type="info", close_button=True, timeout=0,
                )
            release_list.refresh()
        except Exception as exc:
            releases = "error"
            release_list.refresh()
            error("Requesting version information from github.com failed", exc)

    with ui.card().classes("w-full"):
        release_list()

    with ui.row():
        ui.button("Check now", icon="sync", on_click=check, color="positive")
        auto_check = ui.switch("Check on start", value=False).bind_value(app.storage.user, "version_history_autocheck")

    if auto_check.value:
        global releases
        if not releases:  # don't check if we already have data
            releases = []
            ui.timer(0.001, check, once=True)  # Use a timer to detect when the client connects

version_tab = GUITab(
    name="version_history",
    label="Version History",
    icon="update",
    content_func=_version_tab,
)


from argparse import ArgumentParser, RawDescriptionHelpFormatter
import itertools
from json import JSONDecodeError
from pathlib import Path

import numpy as np

from . import synth_format, rails, pattern_generation, movement, utils, __version__

_filter_groups = {
    "notes": synth_format.NOTE_TYPES,
    "walls": list(synth_format.WALL_TYPES),
    "slides": synth_format.SLIDE_TYPES,
}

def _movement_helper(data: synth_format.DataContainer, mirror_left: bool, base_func, relative_func, pivot_func, relative: bool, pivot: list[int], *args, **kwargs) -> None:
    """pick the right function depending on relative or pivot being set"""
    if relative:
        data.apply_for_all(relative_func, *args, mirror_left=mirror_left,  **kwargs)
    elif pivot is not None:
        data.apply_for_all(pivot_func, *args, mirror_left=mirror_left, pivot_3d=pivot, **kwargs)
    else:
        data.apply_for_all(base_func, *args, mirror_left=mirror_left, **kwargs)

def do_movement(options, data: synth_format.DataContainer, filter_types: tuple[str, ...] = synth_format.ALL_TYPES) -> None:
    common_args = dict(relative=options.relative, pivot=np.array(options.pivot), mirror_left=options.mirror_left, types=filter_types)
    if options.scale is not None:
        data.apply_for_all(movement.scale, scale_3d=options.scale, **common_args)
    if options.rotate is not None:
        data.apply_for_all(movement.rotate, angle=options.rotate, **common_args)
    if options.wall_rotate is not None:
        data.apply_for_walls(movement.rotate, angle=options.wall_rotate, relative=True, mirror_left=options.mirror_left, types=filter_types)
    if options.offset is not None:
        data.apply_for_all(movement.offset, offset_3d=options.offset, **common_args)
    if options.outset is not None:
        data.apply_for_all(movement.outset, outset_scalar=options.outset, **common_args)

def do_random_movement(options, data: synth_format.DataContainer, filter_types: tuple[str, ...] = synth_format.ALL_TYPES) -> None:
    if options.rotate_random is not None:
        if len(options.rotate_random) == 1:
            area = np.array(options.rotate_random[0])
        else:
            areas = np.array([
                max(a[1]-a[1], 0.01)  # area, where 0-width areas are counted as 0.01 for numerical stability
                for a in options.rotate_random
            ])
            area = options.rotate_random[np.random.choice(len(areas), p=areas/sum(areas))]
        random_angle = np.random.random_sample() * (area[1]-area[0]) + area[0]
        data.apply_for_all(movement.rotate, angle=random_angle, relative=options.relative, pivot=np.array(options.pivot), mirror_left=options.mirror_left, types=filter_types)

    if options.offset_random is not None:
        if len(options.offset_random) == 1:
            area = options.offset_random[0]
        else:
            areas = np.array([
                max(a[1,0]-a[0,0], 0.01)*max(a[1,1]-a[0,1], 0.01)  # area, where 0-width axes are counted as 0.01 for numerical stability
                for a in options.offset_random
            ])
            area = options.offset_random[np.random.choice(len(areas), p=areas/sum(areas))]
        random_offset = pattern_generation.random_xy(1, area[0], area[1])[0]
        data.apply_for_all(movement.offset, [random_offset[0], random_offset[1], 0], mirror_left=options.mirror_left, types=filter_types)
    
def get_parser():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        prog=f"python3 -m {__package__}.{Path(__file__).stem}",
        description='\n'.join([
            f"Note types:\n\t{', '.join(synth_format.NOTE_TYPES)}",
            f"Wall types:\n\t{', '.join(synth_format.WALL_TYPES)}",
            "",
            "Most number values accept decimals, percentages and fractions (ie '0.25', '25%' or '1/4')",
            "\tTime inputs may also be given in seconds (ie '81.5s' or '1:21.5'), which will be converted into beats",
            "If your value starts with a '-', you must add a = between option and value, ie '--rotate=-45' or '--offset=-1,0,0'",
            "For most options, only the last occurence is considered (exception: '--offset-random' and '--rotate-random')",
            "",
            "Angles are in degrees",
            "Vectors are specified as 'x,y,time':",
            "\tX/Y is measured in editor grid squares, where +x is right and +y is up",
            "\tTime is measured in measures/beats from the start of the selection",
            "",
            "Also see the wiki on GitHub, which contains more detailed explainations, as well as some examples and images: https://github.com/adosikas/synth_mapping_helper/wiki",
        ]),
        epilog=f"Version: {__version__}",
    )

    parser.add_argument("-E", "--start-empty", type=utils.parse_number, metavar="BPM", help="Start with empty data instead of reading the clipboard. Requires specifing BPM. Use with --spawn.")
    parser.add_argument("--use-original", action="store_true", help="When calling this multiple times, start over with orignal copied json")
    parser.add_argument("-f", "--filter-types", nargs="+", metavar="FILTER_TYPE", choices=synth_format.ALL_TYPES + tuple(_filter_groups), default=synth_format.ALL_TYPES, help=f"Only affect notes and walls of these types. Multiple types can be specified seperated by spaces, defaults to all")
    parser.add_argument("--invert-filter", action="store_true", help="Invert filter so everything *but* the filter is affected")
    parser.add_argument("--mirror-left", action="store_true", help="Mirror operations for the LEFT hand")

    preproc_group = parser.add_argument_group("pre-processing")
    preproc_group.add_argument("-b", "--bpm", type=utils.parse_number, help="Change BPM without changing timing")
    preproc_group.add_argument("--delete-others", action="store_true", help="Delete everything that doesn't match the filter")
    preproc_group.add_argument("--connect-singles", type=utils.parse_time, metavar="MAX_INTERVAL", help="Replace strings of single notes with rails if they are . Use with'--rails-to-singles=1' if you want to keep the singles")
    preproc_group.add_argument("--merge-rails", type=utils.parse_time, nargs="?", action="append", metavar="MAX_INTERVAL", help="Merge sequential rails. By default only joins rails that start close to where another ends (in X, Y AND time), but with MAX_INTERVAL only the time interval counts")
    preproc_group.add_argument("--swap-hands", action="store_true", help="Swap left and right hand notes.")
    preproc_group.add_argument("-n", "--change-notes", metavar="NEW_NOTE_TYPE", nargs="+", choices=synth_format.NOTE_TYPES, help=f"Change the type/color of notes. Specify multiple to loop over them.")
    preproc_group.add_argument("-w", "--change-walls", metavar="NEW_WALL_TYPE", nargs="+", choices=synth_format.WALL_TYPES, help=f"Change the type of walls. Specify multiple to loop over them.")
    preproc_group.add_argument("--spawn", metavar="OBJECT_TYPE", choices=synth_format.ALL_TYPES, help="Spawn an objects of the given type. Can only spawn one object per call.")
    preproc_group.add_argument("--spawn-location", type=utils.parse_position, help="Location for the spawned object (defaults to 0,0,0)")

    rail_pattern_group = parser.add_argument_group("rail patterns")
    interp_group = rail_pattern_group.add_mutually_exclusive_group()
    interp_group.add_argument("--interpolate", type=utils.parse_time, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating using splines close to the editor. Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
    interp_group.add_argument("--interpolate-linear", type=utils.parse_time, metavar="INTERVAL", help="Subdivide rail into segments of this length in beats, interpolating linearly. Supports fractions. When used with --spiral or --spikes, this is the distance between each nodes")
    rail_pattern_group.add_argument("--shorten-rails", type=utils.parse_time, metavar="DISTANCE", help="Cut some distance from every rail. Supports fractions. When negative, cuts from the start instead of the end")
    rail_pattern_group.add_argument("--start-angle", type=utils.parse_number, default=0.0, metavar="DEGREES", help="Angle of the first node of the spiral in degrees. Default: 0/right")
    rail_pattern_group.add_argument("--radius", type=utils.parse_number, default=1.0, help="Radius of spiral or length of spikes")
    rail_op_group = rail_pattern_group.add_mutually_exclusive_group()
    rail_op_group.add_argument("--spiral", type=utils.parse_number, metavar="NODES_PER_ROT", help="Generate counterclockwise spiral around rails with this number of nodes per full rotation. Supports fractions. 2=zigzag, negative=clockwise")
    rail_op_group.add_argument("--spikes", type=utils.parse_number, metavar="NODES_PER_ROT", help="Generate spikes from rail, either spiraling (see --spiral) or random (when set to 0)")
    rail_pattern_group.add_argument("--spike-width", type=utils.parse_time, default=1/32, help="Width of spike 'base' in beats. Supports fractions. Should not be lower than 1/32 (the default) and should be lower than chosen interpolation interval")

    movement_group = parser.add_argument_group("movement", description="Operation order is always: scale, rotate, offset, outset")
    pivot_group = movement_group.add_mutually_exclusive_group()
    pivot_group.add_argument("-p", "--pivot", type=utils.parse_position, help="Pivot for outset, scale and rotate as x,y,t")
    pivot_group.add_argument("--note-pivot", choices=synth_format.NOTE_TYPES, help="Use position of first matching note as pivot (determined before any operations)")
    pivot_group.add_argument("--relative", action="store_true", help="Use first node of rails as pivot for outset, scale and rotate, and for walls offset because relative based on rotation")
    movement_group.add_argument("-s", "--scale", type=utils.parse_position, help="Scale positions by x,y,t. Use negative values to mirror across axis. Does NOT change the size of walls. Time-Scale of 2 means twice as long, not twice as fast.")
    movement_group.add_argument("-r", "--rotate", type=utils.parse_number, metavar="DEGREES", help="Rotate everything counterclockwise by this many degrees (negative for clockwise)")
    movement_group.add_argument("--wall-rotate", type=utils.parse_number, metavar="DEGREES", help="Rotate walls counterclockwise around themselves by this many degrees (negative for clockwise)")
    movement_group.add_argument("-o", "--offset", type=utils.parse_position, help="Move/Translate by x,y,t")
    movement_group.add_argument("--outset", type=utils.parse_number, metavar="DISTANCE", help="Move outwards (negative for inwards, can move 'across' pivot)")
    rail_stack_group = movement_group.add_mutually_exclusive_group()
    rail_stack_group.add_argument("--offset-along", choices=synth_format.NOTE_TYPES, help="While stacking: Offset objects to follow notes/rails of the specified type")
    rail_stack_group.add_argument("--rotate-with", choices=synth_format.NOTE_TYPES, help="While stacking: Rotate and outset the objects to follow notes/rails of the specified type")
    movement_group.add_argument("--offset-random", nargs="+", action="extend", type=utils.parse_xy_range, metavar="[MIN_X:]MAX_X,[MIN_Y:]MAX_Y", help="Offset by a random amount in the X and Y axis (ignoring relative & pivot). When no MIN is given, uses negative MAX. Multiple ranges can be specified, separated via space or as separate options (required when starting with '-').")
    movement_group.add_argument("--rotate-random", nargs="+", action="extend", type=utils.parse_range, metavar="[MIN_ANG:]MAX_ANG", help="Rotate by a random angle (obeying relative & pivot if given). When no MIN is given, uses negative MAX. Multiple ranges can be specified, separated via space or as separate options (required when starting with '-').")

    movement_group.add_argument("-c", "--stack-count", type=int, help="Instead of moving, create copies. Must have time offset set.")
    movement_group.add_argument("--stack-duration", type=utils.parse_time, help="Like --stack-count, but you can give it during in beats ('4') or seconds ('2s').")
    movement_group.add_argument("--autostack", nargs="?", choices=("OFFSET", "SPIRAL", "OUTSET", "SCALE"), action="append", metavar="OFFSET/SPIRAL/OUTSET/SCALE", help="Find the first pair of matching objects and continue the pattern. For continuing the positions, the following modes are supported: OFFSET (default, absolute xy), SPIRAL (rotate around implied pivot), OUTSET (distance from pivot) and SCALE (xy ratio).")

    postproc_group = parser.add_argument_group("post-processing")
    postproc_group.add_argument("--parallels", type=utils.parse_number, metavar="DISTANCE", help="Create parallel left/right handed patterns with given spacing from input. Negative numbers result in crossovers.")
    postproc_group.add_argument("--split-rails", action="store_true", help="Split rails at single notes")
    postproc_group.add_argument("--rails-to-singles", type=int, nargs="?", action="append", metavar="KEEP_RAIL", help="Replace rails with single notes at all nodes. KEEP_RAIL is optional and can be '1' if you want to keep the rail instead of replacing it")
    postproc_group.add_argument("--snap-singles-to-rail", action="store_true", help="Snap single notes to rail of the same color")
    postproc_group.add_argument("--keep-alignment", action="store_true", help="Do NOT shift the start of selection to first element")

    return parser

def abort(reason: str):
    if __name__ == "__main__":
        print("ERROR: " + reason)
        exit(1)
    else:
        raise RuntimeError(reason)

def main(options):
    if options.start_empty is not None:
        if options.start_empty <= 0:
            abort("BPM must be positive")
        data = synth_format.DataContainer(options.start_empty, {}, {}, {}, {}, {})
    else:
        try:
            data = synth_format.import_clipboard(options.use_original)
        except (JSONDecodeError, KeyError) as err:
            abort(f"Could not decode clipboard, did you copy somethinge else?\n\t{err!r}")
        except synth_format.JSONParseError as jpe:
            abort(
                "Detected invalid map data. The editor or your map may be corrupted.\n"
                "\tYou can try repairing the file using 'File Utils' tab of the GUI, or restoring from a backup.\n"
                "\tIf you know how to fix the json manually, the following should help (otherwise you may find help on the discord):\n"
                f"\t{jpe!r}\n"
                f"\tCaused by {jpe.__cause__!r}"
            )

    # argument post-parsing
    # convert time parsed in seconds to beats
    for pos_val in "spawn-location", "pivot", "offset", "scale":
        v = getattr(options, pos_val, None)
        if v is not None and isinstance(v[2], utils.SecondFloat):
            setattr(options, pos_val[:2] + (v[2].to_beat(data.bpm),))
    for time_val in "connect_singles", "merge_rails", "interpolate", "interpolate_linear", "shorten_rails", "spike_width", "stack_duration":
        v = getattr(options, time_val, None)
        if isinstance(v, utils.SecondFloat):
            setattr(options, time_val, v.to_beat(data.bpm))
    # expand filter groupts
    if any(name in options.filter_types for name in _filter_groups):
        out_filters = []
        for inp in options.filter_types:
            if inp in _filter_groups:
                out_filters.extend(_filter_groups[inp])
            else:
                out_filters.append(inp)
        options.filter_types = out_filters
    # normalize/invert filter
    if not options.invert_filter:
        # only have each entry once
        filter_types = list(set(options.filter_types))
    else:
        filter_types = list(
            t for t in synth_format.ALL_TYPES
            if t not in options.filter_types
        )
    # get note pivot (before filtering)
    if options.note_pivot:
        notes = getattr(data, options.note_pivot)
        if not notes:
            abort(f"Could not find any {options.note_pivot} notes")
        options.pivot = notes[sorted(notes)[0]][:3]

    # stacking
    if options.autostack is not None:
        if not options.stack_count and not options.stack_duration:
            abort("Cannot --autostack without count/duration")
        obj_type = None
        first = None
        second = None
        for t in synth_format.ALL_TYPES[::-1]:  # start with walls (for SPIRAL mode)
            obj_dict = data.get_object_dict(t)
            if len(obj_dict) >= 2:
                ty_objs = sorted(obj_dict.items())
                _, ty_first = ty_objs[0]
                if first is None or ty_first[0,2] < first[2]:
                    first = ty_first[0]  # rail head only
                    _, ty_second = ty_objs[1]
                    second = ty_second[0]
                    obj_type = t
        if first is None:
            abort("Could not find object pair for autostack")

        if options.autostack[0] == "SPIRAL":
            if first.shape[0] != 5:
                abort("Can only spiral-autostack based on walls")
            # rotation from point 'a' to point 'b', around pivot 'p' by angle 'ang':
            #   x_b = (x_a - x_p) * cos(ang) - (y_a - y_p) * sin(ang) + x_p
            #   y_b = (x_a - x_p) * sin(ang) + (y_a - y_p) * cos(ang) + y_p
            # I couldn't find a good way to solve that for x_p and y_p, so lets do it the geometric way:
            #   p must be an equal distance from both points, so together with p they form an isosceles triangle
            #   In this triangle we want the the angle at p (alpha) to match wall rotation, allowing us to calculate p using basic trigonometry
            #   In particular, the first point, the point halfway between both points and the pivot form a right triangle with an angle of alpha/2 at the pivot and an opposite of (b-a)/2
            #   Therefore, the hypotenuse will be ((b-a)/2) / sin(alpha/2) long and face the direction of (b-a)/2 rotated by 90-alpha/2
            options.rotate = second[4] - first[4]
            options.pivot = first[:3] + movement.rotate((second - first)[:3]/2, 90 - options.rotate/2) / np.sin(np.radians(options.rotate/2))
            options.offset = (0, 0, second[2] - first[2])
        else:
            if options.pivot is not None:
                first[:2] -= options.pivot[:2]
                second[:2] -= options.pivot[:2]

            if options.autostack[0] is None or options.autostack[0] == "OFFSET":
                options.offset = (second - first)[:3]
            elif options.autostack[0] == "OUTSET":
                # this is equivalent to the "rotate-with" calculations below
                options.rotate = np.degrees(np.arctan2(first[0], first[1])) - np.degrees(np.arctan2(second[0], second[1]))
                options.outset = np.sqrt(second[:2].dot(second[:2])) - np.sqrt(first[:2].dot(first[:2]))
                options.offset = (0, 0, second[2] - first[2])
            elif options.autostack[0] == "SCALE":
                options.scale = (second[0] / first[0], second[1] / first[1], 1)
                options.offset = (0, 0, second[2] - first[2])

            if first.shape[0] == 5:
                options.wall_rotate = ((second[4] - first[4]) - (options.rotate or 0)) or None  # see if we need to correct wall rotation

        # remove second object so it doesn't get stacked
        if obj_type in synth_format.NOTE_TYPES:
            del getattr(data, obj_type)[second[2]]
        else:
            del data.walls[second[2]]

    if options.stack_duration:
        if options.stack_count:
            abort("Cannot use --stack-count and --stack-duration at the same time")
        options.stack_count = int(options.stack_duration / options.offset[2])

    if options.stack_count and (options.offset is None or not options.offset[2]):
        abort("Cannot stack with time offset of 0")

    # preprocessing
    if options.bpm:
        bpm_scale = [1, 1, options.bpm / data.bpm]
        data.apply_for_all(movement.scale, bpm_scale, types=filter_types)

    if options.delete_others:
        data = data.filtered(types=filter_types)
    if options.connect_singles:
        data.apply_for_note_types(rails.connect_singles, max_interval=options.connect_singles, types=filter_types)
    if options.merge_rails:
        if options.merge_rails[0] is None:
            data.apply_for_note_types(rails.merge_sequential_rails, types=filter_types)
        else:
            data.apply_for_note_types(rails.merge_rails, max_interval=options.merge_rails[0], types=filter_types)

    if options.swap_hands:
        data.left, data.right = data.right, data.left

    if options.change_notes:
        if len(options.change_notes) == 1:
            # to single type: just merge all dicts
            changed = {}
            for t in filter_types:
                if t in synth_format.NOTE_TYPES and t != options.change_notes[0]:
                    changed |= getattr(data, t)
                    setattr(data, t, {})
            # existing notes always have priority
            setattr(data, options.change_notes[0], changed | getattr(data, options.change_notes[0]))
        else:
            # to multiple types: cycle
            outputs = {}
            for t in options.change_notes:
                outputs[t] = {}
            for t in filter_types:
                if t in synth_format.NOTE_TYPES:
                    notes = getattr(data, t)
                    # note: looping happens independently for each note type
                    for time, new_type in zip(sorted(notes), itertools.cycle(options.change_notes)):
                        nodes = notes[time]
                        outputs[new_type][nodes[0,2]] = nodes
                    setattr(data, t, {})  # clear existing notes
            for t, notes in outputs.items():
                # existing notes (that did not get changed) always have priority
                setattr(data, t, notes | getattr(data, t))

    if options.change_walls:
        if len(options.change_walls) == 1:
            # to single type: just merge all arrays the same way
            new_type = synth_format.WALL_TYPES[options.change_walls[0]][0]
            data.apply_for_walls(pattern_generation.change_wall_type, types=filter_types, new_type=new_type)
        else:
            # to multiple types: cycle
            generator = itertools.cycle(synth_format.WALL_TYPES[t][0] for t in options.change_walls)
            def _change_wall_type(wall: "numpy array (1, 5)", direction: int=1) -> "numpy array (1, 5)":
                new_wall = wall.copy()
                new_wall[..., 3] = next(generator)
                return new_wall
            data.apply_for_walls(_change_wall_type, types=filter_types)

    if options.spawn:
        loc = options.spawn_location or (0.0,0.0,0.0)
        if options.spawn in synth_format.NOTE_TYPES:
            getattr(data, options.spawn)[loc[2]] = np.array([loc])
        elif options.spawn in synth_format.WALL_TYPES:
            data.walls[loc[2]] = np.array([loc + (synth_format.WALL_TYPES[options.spawn][0],0)])

    # rail patterns
    if options.interpolate:
        data.apply_for_notes(rails.interpolate_nodes, "spline", options.interpolate, types=filter_types)
    elif options.interpolate_linear:
        data.apply_for_notes(rails.interpolate_nodes, "linear", options.interpolate_linear, types=filter_types)
    if options.shorten_rails:
        data.apply_for_notes(rails.shorten_rail_by, options.shorten_rails, types=filter_types)
    if options.spiral:
        if (1 / options.spiral) % 1 == 0:
            abort("Chosen spiral factor divides 1 and would result in a straight rail. Refusing action!")
        data.apply_for_notes(
            pattern_generation.add_spiral,
            fidelity=options.spiral,
            radius=options.radius,
            start_angle=options.start_angle,
            types=filter_types,
            mirror_left=options.mirror_left,
        )
    if options.spikes is not None:
        data.apply_for_notes(
            pattern_generation.add_spikes,
            fidelity=options.spikes,
            radius=options.radius,
            spike_duration=options.spike_width,
            start_angle=options.start_angle,
            types=filter_types,
            mirror_left=options.mirror_left,
        )

    # movement
    if options.stack_count:
        stacking = data.filtered(types=filter_types)  # pre-filter
        if options.offset_along or options.rotate_with:
            if options.offset_along and options.rotate_with:
                abort("Cannot use offset-along and rotate-with at the same time")
            if options.relative:
                abort("Cannot use offset-along or rotate-with in relative mode")
            start_pos = rails.get_position_at(getattr(data, options.offset_along or options.rotate_with), 0)
            if start_pos is None:
                abort(f"No start position of {options.offset_along or options.rotate_with} notes found")
            if options.pivot is not None:
                start_pos -= options.pivot[:2]
        for i in range(options.stack_count):
            do_movement(options, stacking)
            tmp = stacking.filtered()  # deep copy
            if options.offset_along:
                cur_pos = rails.get_position_at(getattr(data, options.offset_along), (i+1) * options.offset[2])
                if cur_pos is None:
                    abort(f"No intermediate position of {options.offset_along} notes found at {(i+1) * options.offset[2]}")
                delta = cur_pos - start_pos
                tmp.apply_for_all(movement.offset, [delta[0], delta[1], 0], mirror_left=options.mirror_left)
            if options.rotate_with:
                cur_pos = rails.get_position_at(getattr(data, options.rotate_with), (i+1) * options.offset[2])
                if cur_pos is None:
                    abort(f"No intermediate position of {options.rotate_with} notes found at {(i+1) * options.offset[2]}")
                if options.pivot is not None:
                    cur_pos -= options.pivot[:2]
                angle_diff = np.degrees(np.arctan2(start_pos[0], start_pos[1])) - np.degrees(np.arctan2(cur_pos[0], cur_pos[1]))
                distance_diff = np.sqrt(cur_pos.dot(cur_pos)) - np.sqrt(start_pos.dot(start_pos))
                tmp.apply_for_all(movement.rotate, angle=angle_diff, pivot=np.array(options.pivot), mirror_left=options.mirror_left)
                # outset all objects together by precalculating the offset based on rail position
                group_outset_offset = [cur_pos[0], cur_pos[1], 0] / np.sqrt(cur_pos.dot(cur_pos)) * distance_diff
                tmp.apply_for_all(movement.offset, offset_3d=group_outset_offset) 
            do_random_movement(options, tmp)
            data.merge(tmp)
    else:
        if options.offset_along or options.rotate_with:
            abort("Cannot use offset-along or rotate-with without stacking")
        do_movement(options, data, filter_types=filter_types)
        do_random_movement(options, data, filter_types=filter_types)

    # postprocessing
    if options.parallels:
        pattern_generation.create_parallel(data, options.parallels)
    if options.split_rails:
        data.apply_for_note_types(rails.split_rails, types=filter_types)
    if options.rails_to_singles:
        data.apply_for_note_types(rails.rails_to_singles, keep_rail=bool(options.rails_to_singles[0]), types=filter_types)
    if options.snap_singles_to_rail:
        data.apply_for_note_types(rails.snap_singles_to_rail, types=filter_types)
    synth_format.export_clipboard(data, not options.keep_alignment)

def entrypoint():
    main(get_parser().parse_args())

if __name__ == "__main__":
    entrypoint()



from argparse import ArgumentError
import shlex

from nicegui import app, events, ui

from .utils import *
from .. import cli

DEFAULT_PRESETS = {
    "Merge Rails": """
        # NOTE: You can do this via the Dashboard too: Merge sequential rails
        --merge-rails""",
    "Split Rails": """
        # NOTE: You can do this via the Dashboard too: Split rails at single notes
        --split-rails""",
    "Spikify": """
        # NOTE: You can do this via the Dashboard too: Spikes, Angle=180
        # add up-down spikes every 1/4 beat to a rail
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Rail-Options#spikes
        --interpolate=1/4 --spikes=2 --radius=2 --start-angle=90""",
    "Spiralize": """
        # NOTE: You can do this via the Dashboard too: Spiral, Angle=45
        # turn a rail into a spiral around the original path, with one rotation per beat and starting top/right
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Rail-Options#spiral
        --interpolate=1/8 --spiral=8 --radius=2 --start-angle=45""",
    "Parallels": """
        # NOTE: You can do this via the Dashboard too: Parallel, Spacing=2
        # create parallel rails to the input, with 2 sq distance. left/right notes are added additionally to the left/right of the right/left notes, single/both notes are "split"
        # There will be conflicts when you have multiple notes or start of rails at the same time. Check the wiki on how it decides on priority or just avoid that.
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Pre--and-Post-Processing-Options#create-parallel-patterns
        --parallels=2""",
    "Autostack circular": """
        # NOTE: You can do this via the Stacking Tab too: Pick button in "Rotatation" card, then stack 15x
        # automatically detect rotation & outset between the first two objects of the same type and continue the pattern for a total of 16 (first + 15 stacked copies)
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Movement-options#autostacking
        --autostack=OUTSET --stack-count=15""",
    "Autostack linear": """
        # NOTE: You can do this via the Stacking Tab too: Pick button in "Offset" card, then stack for 8b
        # automatically detect offset between the first two objects of the same type and continue the pattern for 8 measures
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Movement-options#autostacking
        --autostack=OFFSET --stack-duration=8""",
    "Autostack spiral walls": """
        # NOTE: You can do this via the Stacking Tab too: Pick button in "All" card, then stack for 8b
        # automatically detect spiral shape from first two walls and continue the patterns for 8 measures
        # you can have notes and other walls in the pattern, but ensure the first PAIR of objects is walls
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Movement-options#autostacking
        --autostack=SPIRAL --stack-duration=8""",
    "Rail to notestack": """
        # NOTE: You can do this via the Dashboard too: "Interpolate rail nodes", then "Rail to notestack (delete rail)" (or keep rail)
        # convert rails to notestacks with 1/16 spacing
        # you can also do "--rails-to-singles=1" to KEEP the rail instead of removing it
        # See https://github.com/adosikas/synth_mapping_helper/wiki/Pre--and-Post-Processing-Options#convert-rails-into-single-notes
        --interpolate=1/16 --rails-to-singles""",
    "Oops! All rails": """
        # NOTE: You can do this via the Dashboard too: "Rail nodes to notes (delete rail)", then "Connect notes"
        # convert everything that is no more than 1 beat apart to a single rail 
        --rails-to-singles
        --connect-singles=1""",
    "Stack with green rail": """
        # stack the selection 4 times with 1/4 spacing, rotating with a green (single handed special) rail (but not also stacking that rail itself)
        --offset=0,0,1/4 --stack-count=4 --rotate-with=single  --filter=single --invert-filter
        """,
}

# strip newlines and tabs
DEFAULT_PRESETS = {
    k: '\n'.join(l.strip() for l in v.splitlines() if l)
    for k,v in DEFAULT_PRESETS.items()
}

def _command_tab() -> None:
    presets: dict[str, str] = {}

    loaded_presets = app.storage.user.get("command_presets")
    if loaded_presets:
        logger.info(f"Loaded {len(loaded_presets)} presets")
        presets = loaded_presets
    else:
        logger.info(f"No presets saved, loading defaults")
        presets = DEFAULT_PRESETS.copy()

    @ui.refreshable
    def quick_run_buttons() -> None:
        for p in presets:
            ui.button(p, icon="fast_forward", on_click=run_preset).tooltip(f"Load and immediately execute {p} preset")

    def presets_updated() -> None:
        preset_selector.set_options(list(presets))
        quick_run_buttons.refresh()

    @handle_errors
    def load_commands(e: events.UploadEventArguments) -> None:
        upl: ui.upload = e.sender  # type: ignore
        upl.reset()
        data = e.content.read()
        try:
            presets[e.name] = data.decode()
            presets_updated()
            preset_selector.value = e.name  # this also loads the content
        except UnicodeDecodeError as ude:
            raise PrettyError(msg=f"Error reading commands from {e.name}", exc=ude, data=data[ude.start: ude.end].hex())

    @handle_errors
    def run_command() -> None:
        p = cli.get_parser()
        p.exit_on_error = False
        commands = command_input.value.splitlines()
        count = 0
        for i, line in enumerate(commands):
            if line.startswith("#"):
                continue
            args = shlex.split(line)
            if not count and use_orig_cb.value:
                args.append("--use-original")
            if mirror_left_cb.value:
                args.append("--mirror-left")

            try:
                opts, remaining = p.parse_known_args(args)
            except ArgumentError as ae:
                raise PrettyError(msg=f"Error parsing line {i+1}", exc=ae, data=line) from ae
            if remaining:
                raise PrettyError(msg=f"Unknown arguments in line {i+1}", data=remaining, exc=ValueError(remaining))
            try:
                cli.main(opts)
            except RuntimeError as re:
                raise PrettyError(msg=f"Error running line {i+1}", exc=re, data=line) from re
            except Exception as exc:
                raise PrettyError(msg=f"Unexpected error running line {i+1}. Please report this by clicking the report button in the top-right and sending me the report. ", exc=exc, data=line) from exc
            count += 1
        else:
            if preset_selector.value:
                message = f"Sucessfully executed preset '{preset_selector.value}' ({count} command{'s'*(count>1)})"
            else:
                message = f"Sucessfully executed {count} command{'s'*(count>1)}"
            info(message)

    def run_preset(e: events.ClickEventArguments) -> None:
        btn: ui.button = e.sender  # type: ignore
        preset_selector.value = btn.text
        run_command()

    def restore_presets() -> None:
        presets.update(DEFAULT_PRESETS)
        presets_updated()
        prev_val = preset_selector.value
        preset_selector.value = None
        if prev_val in presets:
            preset_selector.value = prev_val
        info("Restored default presets")

    def save_presets() -> None:
        app.storage.user["command_presets"] = presets
        info(f"Saved {len(presets)} presets")

    def add_preset() -> None:
        presets[add_preset_name.value] = command_input.value
        presets_updated()
        preset_selector.value = add_preset_name.value
        add_dialog.close()

    with ui.dialog() as add_dialog, ui.card():
        add_preset_name = ui.input("Preset name", value="Untitled").props("autofocus")
        with ui.row():
            ui.button("Cancel", icon="cancel", on_click=add_dialog.close).props("flat")
            ui.button("Add", icon="add", color="green", on_click=add_preset)

    def delete_preset(e: events.ClickEventArguments) -> None:
        del presets[preset_selector.value]
        presets_updated()
        preset_selector.value = None
        remove_dialog.close()

    with ui.dialog() as remove_dialog, ui.card():
        remove_confirmation_label = ui.label("Really delete?")
        with ui.row():
            ui.button("Cancel", icon="cancel", on_click=remove_dialog.close).props("flat")
            ui.button("Delete", icon="delete", color="red", on_click=delete_preset)

    with ui.card():
        with ui.row():
            with ui.select(list(presets), label="Preset Name", with_input=True).props("dense") as preset_selector:
                ui.tooltip("Select a preset")
            preset_selector.bind_value(app.storage.user, "command_preset")
            with ui.button(icon="delete", color="red", on_click=remove_dialog.open).classes("my-auto").bind_enabled_from(preset_selector, "value"):
                ui.tooltip("Delete current preset")
            with ui.button(icon="add", color="green", on_click=add_dialog.open).classes("my-auto") as add_button:
                ui.tooltip("Add current command as preset")
            ui.separator().props("vertical")
            with ui.button(icon="save", on_click=save_presets).props("outline").classes("my-auto"):
                ui.tooltip("Store presets")
            ui.upload(label="Import files", auto_upload=True, multiple=True, on_upload=load_commands).props('color="positive" flat').classes("w-40")
            with ui.button(icon="post_add", on_click=restore_presets, color="red").props("outline").classes("my-auto"):
                ui.tooltip("Restore default presets")
        command_input = ui.textarea("Commands", placeholder="--offset=1,0,0").props("outlined").classes("w-full")
        command_input.bind_value(app.storage.user, "command_input")

        # clear preset name when content changes
        def _clear_preset(e: events.ValueChangeEventArguments) -> None:
            if presets.get(preset_selector.value) != e.value:
                preset_selector.set_value(None)
        command_input.on_value_change(_clear_preset)
        # set content when preset changes to a valid preset
        def _load_preset(e: events.ValueChangeEventArguments) -> None:
            preset = e.value
            if preset and preset in presets:
                command_input.set_value(presets[preset])
        preset_selector.on_value_change(_load_preset)
    
        preset_selector.bind_value_to(remove_confirmation_label, "text", forward=lambda v: f"Really delete '{v}'?")
        add_button.bind_enabled_from(command_input, "value")

        with ui.row():
            ui.button("Execute", icon="play_arrow", on_click=run_command).bind_enabled_from(command_input, "value").tooltip("Execute commands and copy result to clipboard")
            with ui.checkbox("Use original JSON") as use_orig_cb:
                wiki_reference("Miscellaneous-Options#use-original-json")
            with ui.checkbox("Mirror for left hand") as mirror_left_cb:
                wiki_reference("Miscellaneous-Options#mirror-operations-for-left-hand")
        ui.separator()
    
        ui.label("Quick run:").tooltip("Stored presets will show here")
        with ui.row():
            quick_run_buttons()

command_tab = GUITab(
    name="command",
    label="Commands",
    icon="play_arrow",
    content_func=_command_tab,
)



#/usr/bin/env python3
import base64
from io import BytesIO
import codecs
from contextlib import contextmanager
import dataclasses
import datetime
from hashlib import sha256
import json
from pathlib import Path
import re
import time
from typing import Any, Generator, Literal, Union
import struct
import sys
import zipfile

import numpy as np
import pyperclip
import soundfile

from synth_mapping_helper import movement, __version__
from synth_mapping_helper.audio_format import AudioData
from synth_mapping_helper.utils import second_to_beat, beat_to_second

# For simplicity, we exclusively use grid coordinates (x, y) and use measures for time (z)
# These values are only needed to convert to / from the format the game uses

INDEX_SCALE = 64  # index = measure * INDEX_SCALE
GRID_SCALE = 0.1365  # xy_coord = xy_grid * GRID_SCALE
TIME_SCALE = 20  # z_coord = second * TIME_SCALE

# Apparently the editor grid snap is off by a small amount...
X_OFFSET = 0.002
Y_OFFSET = 0.0012

NOTE_TYPES = ("right", "left", "single", "both")
NOTE_TYPE_STRINGS = ("RightHanded", "LeftHanded", "OneHandSpecial", "BothHandsSpecial")
# Note: wall offsets are eyeballed
WALL_TYPES: dict[str, tuple[int, list[float]]] = {
    # name: (index, center_offset)
    # index 0-4 are the same "slideType" used in the JSON in the "slides" list
    "wall_right": (0, [4.05, -1.85]),
    "wall_left": (1, [-3.95, -1.85]),
    "angle_right": (2, [2.5, -2.2]),
    "center": (3, [0, -1.3]),
    "angle_left": (4, [-2.5, -2.2]),
    # 5-99: placeholder for future slide types
    # Since there can be only be one wall anyway, just treat these three the same as slides
    "crouch": (100, [0.1, 7.5]),
    "square": (101, [0, -2]),
    "triangle": (102, [0, -4.7]),
}
WALL_LOOKUP = {id: name for name, (id, _) in WALL_TYPES.items()}
WALL_OFFSETS = {id: np.array(offset + [0]) for _, (id, offset) in WALL_TYPES.items()}
WALL_MIRROR_ID = {id: WALL_TYPES[
        name.replace("left", "right") if "left" in name
        else name.replace("right", "left") if "right" in name
        else name
    ][0]
    for name, (id, _) in WALL_TYPES.items()
}
# symmetry angles in degrees, 360=no symmetry
WALL_SYMMETRY = {
    "wall_right": 180,
    "wall_left": 180,
    "angle_right": 360,
    "center": 180,
    "angle_left": 360,
    "crouch": 360,
    "square": 90,
    "triangle": 120,
}
SLIDE_TYPES = tuple(name for name, (id, _) in WALL_TYPES.items() if id < 100)
LEFT_WALLS = [id for name, (id, _) in WALL_TYPES.items() if "left" in name]
WALL_VERTS = {
    "angle_right": (
        (7.4, -8.75), (-4.6, 8.75), (-5.85, 8.75), (-5.85, 0.2), (5.25, -8.75)
    ),
    "square": (  # drawn as single, concave polygon (similar to a "c", but with the ends touching)
        (8.4, 8.4), (8.4, -8.4), (-8.4, -8.4), (-8.4, 8.4),
        (8.4, 8.4),  # complete the outer ring
        (7.5, 7.5), (-7.5, 7.5), (-7.5, -7.5), (7.5, -7.5),  # inner ring in reverse order (going "back")
        (7.5, 7.5)  # complete inner ring
    ),
    "triangle": (
        (0, -11.2), (-9.7, 5.6), (9.7, 5.6),
        (0, -11.2),
        (0, -10), (8.6, 5), (-8.6, 5),
        (0, -10),
    )
}
# mirror x/y:
for w, (x1, y1), (x2, y2) in (
    ("wall_right", (2.8, 8.4), (4, 6)),
    ("center", (0.15, 11), (2.1, 7.5)),
    ("crouch", (5.6, 5.6), (8, 4)),
):
    WALL_VERTS[w] = ((x1, y1), (x2, y2), (x2, -y2), (x1, -y1), (-x1, -y1), (-x2, -y2), (-x2, y2), (-x1, y1))
# mirror _right to _left:
for w in ("wall_", "angle_"):
    WALL_VERTS[w + "left"] = tuple((-x, y) for x, y in WALL_VERTS[w + "right"][::-1])

ALL_TYPES = NOTE_TYPES + tuple(WALL_TYPES) + ("lights", "effects")

SINGLE_COLOR_NOTES = dict[float, "numpy array (n, 3)"]   # rail segment (n>1) and x,y,t
WALLS = dict[float, "numpy array (1, 5)"]    # x,y,t, type, angle

BEATMAP_JSON_FILE = "beatmap.meta.bin"
METADATA_JSON_FILE = "track.data.json"
DIFFICULTIES = ("Easy", "Normal", "Hard", "Expert", "Master", "Custom")
DIFFICULTY_SPEEDS = {
    "Easy": 1,
    "Normal": 1,
    "Hard": 1,
    "Export": 1.25,
    "Master": 1.5,
}
META_KEYS = ("Name", "Author", "Beatmapper", "CustomDifficultyName", "BPM")

BETA_WARNING_SHOWN = False

class JSONParseError(ValueError):
    def __init__(self, d: dict|list, t: str) -> None:
        super().__init__()
        self.dict = d
        self.type = t

    def __str__(self) -> str:
        return f"Error while parsing JSON of {self.type}: {self.dict}"

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"


def beta_warning() -> None:
    global BETA_WARNING_SHOWN
    if not BETA_WARNING_SHOWN:
        print("\n\n ⚠️  W A R N I N G ⚠️")
        print("\tThis was tested with the beta version of the editor only.")
        print("\tYou may want to switch to it if you encounter bugs or weird behavior.\n\n")
        BETA_WARNING_SHOWN = True
        time.sleep(0.5)

def round_time_to_fractions(time: float) -> float:
    # Round to 1/64 or 1/48, whichever is closer
    r64 = round(time * 64) / 64
    r48 = round(time * 48) / 48
    return r64 if abs(time-r64) <= abs(time-r48) else round(r48, 11)

def round_tick_for_json(time: float) -> float:
    # same as above, but in 1/64 ticks the json needs for some things
    # this is a seperate function to only do one float operation after rounding before output, to minimize errors
    r64 = float(round(time))
    r48 = round(time * 0.75) / 0.75  # 0.75 = 48/64
    return r64 if abs(time-r64) <= abs(time-r48) else round(r48, 11)

def round_tick_for_json_index(time: float) -> float | int:
    # same as above, but for notes dict, which uses int when possible
    if time.is_integer():
        return int(time)
    return round_tick_for_json(time)

# basic coordinate
def coord_from_synth(bpm: float, startMeasure: float, coord: list[float], c_type: str = "unlabeled") -> "numpy array (3)":
    try:
        return np.array([
            (coord[0] - X_OFFSET) / GRID_SCALE,
            (coord[1] - Y_OFFSET) / GRID_SCALE,
            # convert absolute coordinate to number of beats since start
            round_time_to_fractions(second_to_beat(coord[2] / TIME_SCALE, bpm) - startMeasure / 64),
        ])
    except (ValueError, TypeError) as exc:
        raise JSONParseError(coord, c_type) from exc

def coord_to_synth(bpm: float, coord: "numpy array (3)") -> list[float]:
    return [
        round((coord[0] * GRID_SCALE) + X_OFFSET, 11),
        round((coord[1] * GRID_SCALE) + Y_OFFSET, 11),
        round(beat_to_second(round_time_to_fractions(coord[2]), bpm) * TIME_SCALE, 11),
    ]

# full note dict
def note_from_synth(bpm: float, startMeasure: float, note_dict: dict) -> tuple[int, "numpy array (n, 3)"]:
    start = coord_from_synth(bpm, startMeasure, note_dict["Position"], "note" if note_dict["Segments"] is None else "rail head")
    note_type = note_dict["Type"]
    if note_dict["Segments"] is None:
        return note_type, start[np.newaxis]  # just add new axis
    else:
        return note_type, np.stack((start,) + tuple(coord_from_synth(
            bpm, startMeasure, node, f"rail node #{i} of rail starting at json index {startMeasure + start[2]*64}"
        ) for i, node in enumerate(note_dict["Segments"])))

def note_to_synth(bpm: float, note_type: int, nodes: "numpy array (n, 3)") -> dict:
    return {
        "Position": coord_to_synth(bpm, nodes[0]),
        "Segments": [coord_to_synth(bpm, node) for node in nodes[1:]] if nodes.shape[0] > 1 else None,
        "Type": note_type,
    }

# full wall dict
def wall_from_synth(bpm: float, startMeasure: float, wall_dict: dict, wall_type: int) -> "numpy array (1, 5)":
    if wall_type not in WALL_LOOKUP:
        raise JSONParseError(wall_dict, "wall") from ValueError(f"Unexpected wall type ({wall_type})")
    return np.concatenate((
        coord_from_synth(bpm, startMeasure, wall_dict["position"], f"{WALL_LOOKUP[wall_type]} wall") + WALL_OFFSETS[wall_type],
        (wall_type, wall_dict.get("zRotation", 0.0))
    ))[np.newaxis]

def wall_to_synth(bpm: float, wall: "numpy array (1, 5)") -> tuple[str, dict]:
    wall_type = int(wall[0, 3])
    pos = coord_to_synth(bpm, wall[0, :3] - WALL_OFFSETS[wall_type])
    wall_dict = {
        "time": round_tick_for_json(wall[0, 2] * 64),  # time as 1/64
        "slideType": wall_type,
        "position": pos,
        "zRotation": round(wall[0, 4] % 360, 3),  # note: crouch walls cannot be rotated, this will be ignored for them
        "initialized": True,  # no idea what this is for
    }
    # crouch, square and triangle are not in the "slides" list, each has their own list and they do not use the "slideType" key
    if wall_type < 100:  # we gave crouch, square and triangle the types 100, 101 and 102
        dest_list = "slides"
    else:
        dest_list = WALL_LOOKUP[wall_type] + "s"
        del wall_dict["slideType"]
        if wall_type == WALL_TYPES["crouch"][0]:
            # cannot rotate crouch walls
            del wall_dict["zRotation"]
    return dest_list, wall_dict

@dataclasses.dataclass
class RailFilter:
    single: bool = True
    rails: bool = True
    min_len: float | None = None
    max_len: float | None = None
    min_count: int | None = None
    max_count: int | None = None
    min_spacing: float | None = None
    max_spacing: float | None = None

    def matches(self, nodes: "numpy array (n, 3)") -> bool:
        if nodes.shape[0] == 1:
            return self.single
        if not self.rails:
            return False
        if self.min_count is not None and nodes.shape[0] < self.min_count:
            return False
        if self.max_count is not None and nodes.shape[0]  > self.max_count:
            return False
        l = nodes[-1,2] - nodes[0,2]
        if self.min_len is not None and l < self.min_len:
            return False
        if self.max_len is not None and l > self.max_len:
            return False
        spacing = np.diff(nodes[:,2])
        if self.min_spacing is not None and spacing.min() < self.min_spacing:
            return False
        if self.max_spacing is not None and spacing.max() > self.max_spacing:
            return False
        return True

    def __bool__(self) -> bool:
        return (
            not self.single
            or not self.rails
            or self.min_len is not None
            or self.max_len is not None
            or self.min_spacing is not None
            or self.max_spacing is not None
        )

@dataclasses.dataclass
class DataContainer:
    bpm: float = 60.0
    right: SINGLE_COLOR_NOTES = dataclasses.field(default_factory=dict)
    left: SINGLE_COLOR_NOTES = dataclasses.field(default_factory=dict)
    single: SINGLE_COLOR_NOTES = dataclasses.field(default_factory=dict)
    both: SINGLE_COLOR_NOTES = dataclasses.field(default_factory=dict)
    walls: WALLS = dataclasses.field(default_factory=dict)
    # reuse same structure as notes for common code
    lights: SINGLE_COLOR_NOTES = dataclasses.field(default_factory=dict)
    effects: SINGLE_COLOR_NOTES = dataclasses.field(default_factory=dict)

    def __bool__(self) -> bool:
        # truthy when there is *any* object in any of the dicts
        return any(bool(getattr(self, f)) for f in NOTE_TYPES + ("walls", "lights", "effects"))

    def get_counts(self) -> dict[str, dict[str, int]|int]:
        out: dict[str, dict[str, int]] = {
            "walls": {k: 0 for k in WALL_TYPES},
            "notes": {},
            "rails": {},
            "rail_nodes": {},
        }
        for wall in self.walls.values():
            out["walls"][WALL_LOOKUP[wall[0, 3]]] += 1
        for t in NOTE_TYPES:
            n_nodes = [n.shape[0] for n in getattr(self, t).values()]
            notes = sum(1 if n==1 else 0 for n in n_nodes)
            out["notes"][t] = notes
            out["rails"][t] = len(n_nodes) - notes
            out["rail_nodes"][t] = sum(n_nodes) - len(n_nodes)
        for e in ("walls", "notes", "rails", "rail_nodes"):
            out[e]["total"] = sum(out[e].values())
        return out | {"lights": len(self.lights), "effects": len(self.effects)}

    # Note: None of these functions are allowed to *modify* the dicts, instead they must create new dicts
    # This avoids requring deep copies for everything

    def apply_for_notes(self, f, *args, types: tuple[str, ...] = NOTE_TYPES, rail_filter: RailFilter | None = None, mirror_left: bool = False, **kwargs) -> None:
        for t in NOTE_TYPES:
            if t not in types:
                continue
            notes = getattr(self, t)
            out = {}
            for _, nodes in sorted(notes.items()):
                if not rail_filter or rail_filter.matches(nodes):
                    out_nodes = f(nodes, *args, direction=(-1 if mirror_left and t == "left" else 1), **kwargs)
                    out[out_nodes[0, 2]] = out_nodes
                else:
                    out[nodes[0, 2]] = nodes
            setattr(self, str(t), out)

    def apply_for_walls(self, f, *args, types: tuple[str, ...] = tuple(WALL_TYPES), rail_filter: RailFilter | None = None, mirror_left: bool = False, **kwargs) -> None:
        wall_types = [WALL_TYPES[t][0] for t in WALL_TYPES if t in types]
        out_walls = {}
        for time_index, wall in sorted(self.walls.items()):
            if wall[0, 3] in wall_types:
                wall = f(wall, *args, direction=(-1 if mirror_left and wall[0, 3] in LEFT_WALLS else 1), **kwargs)
                out_walls[wall[0, 2]] = wall
            else:
                out_walls[time_index] = wall
        self.walls = out_walls

    def apply_for_all(self, f, *args, types: tuple[str, ...] = ALL_TYPES, rail_filter: RailFilter | None = None, mirror_left: bool = False, **kwargs) -> None:
        self.apply_for_notes(f, *args, types=types, mirror_left=mirror_left, rail_filter=rail_filter, **kwargs)
        self.apply_for_walls(f, *args, types=types, mirror_left=mirror_left, **kwargs)
        for t in ("lights", "effects"):
            if t not in types:
                continue
            objs = getattr(self, t)
            out = {}
            for _, nodes in sorted(objs.items()):
                out_nodes = f(nodes, *args, **kwargs)
                out[out_nodes[0, 2]] = out_nodes
            setattr(self, t, out)


    # used when the functions needs access to all notes and rails of a color at once
    def apply_for_note_types(self, f, *args, types: tuple[str, ...] = NOTE_TYPES, rail_filter: RailFilter | None = None, mirror_left: bool = False, **kwargs) -> None:
        for t in types:
            if t not in NOTE_TYPES:
                continue
            if not rail_filter:
                setattr(self, t, f(getattr(self, t), *args, direction=(-1 if mirror_left and t == "left" else 1), **kwargs))
            else:
                # sort into affected and unaffected
                affected: SINGLE_COLOR_NOTES = {}
                unaffected: SINGLE_COLOR_NOTES = {}
                for ti, nodes in getattr(self, t).items():
                    (affected if rail_filter.matches(nodes) else unaffected)[ti] = nodes
                # output unaffected+modified (modified takes priority)
                setattr(self, t, unaffected | f(affected, *args, direction=(-1 if mirror_left and t == "left" else 1), **kwargs))

    def filtered(self, types: tuple[str, ...] = ALL_TYPES, rail_filter: RailFilter | None = None) -> "DataContainer":
        new_notes: dict[str, SINGLE_COLOR_NOTES] = {}
        for t in NOTE_TYPES:
            if t not in types:
                new_notes[t] = {}
            elif not rail_filter:
                new_notes[t] = getattr(self, t)
            else:
                new_notes[t] = {
                    ti: nodes
                    for ti, nodes in getattr(self, t).items()
                    if rail_filter.matches(nodes)
                }
        wall_types = [WALL_TYPES[t][0] for t in types if t in WALL_TYPES]
        new_walls = {
            time_index: wall
            for time_index, wall in sorted(self.walls.items())
            if wall[0, 3] in wall_types
        }
        new_lights = {} if "lights" not in types else self.lights
        new_effects = {} if "effects" not in types else self.effects
        return dataclasses.replace(self, **new_notes, walls=new_walls, lights=new_lights, effects=new_effects)
        
    def merge(self, other: "DataContainer") -> None:
        for t in NOTE_TYPES:
            self_notes = getattr(self, t)
            other_notes = getattr(other, t)
            setattr(self, t, self_notes | other_notes)
        self.walls |= other.walls
        self.lights |= other.lights
        self.effects |= other.effects

    def get_object_dict(self, type_name: str) -> Union[SINGLE_COLOR_NOTES, WALLS]:
        if type_name in NOTE_TYPES + ("lights", "effects"):
            return getattr(self, type_name)
        wall_type = WALL_TYPES[type_name][0]
        return {
            time_index: wall
            for time_index, wall in sorted(self.walls.items())
            if wall[0, 3] == wall_type
        }

    def find_first(self, types: tuple[str, ...] = ALL_TYPES, rail_filter: RailFilter | None = None) -> tuple[str, "numpy array (3+)"] | None:
        # type, data
        first: tuple[str, "numpy array (3+)"] | None = None
        for t in types:
            if t not in NOTE_TYPES or not rail_filter:
                # rail filter only affects note types
                ty_objs = sorted(self.get_object_dict(t).items())
            else:
                ty_objs = sorted((ti, nodes) for ti, nodes in self.get_object_dict(t).items() if rail_filter.matches(nodes))
            if not ty_objs:
                continue
            _, ty_first = ty_objs[0]
            if first is None or ty_first[0,2] < first[1][2]:
                first = t, ty_first[0]
        return first

    def to_clipboard_json(self, realign_start: bool = True) -> str:
        clipboard: dict[str, Any] = {
            "BPM": self.bpm,
            "startMeasure": 0,
            "startTime": 0,
            "lenght": 0,
            "notes": {},
            "effects": [],
            "jumps": [],
            "crouchs": [],
            "squares": [],
            "triangles": [],
            "slides": [],
            "lights": [],
        }
        if isinstance(self, ClipboardDataContainer) and self.original_json:
            clipboard["original_json"] = self.original_json

        first = None
        last = None
        for note_type, notes in enumerate((self.right, self.left, self.single, self.both)):
            type_notes = {}  # buffer single type to avoid multiple of the same type at the same time
            for time_index, nodes in notes.items(): 
                type_notes[round_tick_for_json(time_index * 64)] = note_to_synth(self.bpm, note_type, nodes)
                if first is None or nodes[0, 2] < first:
                    first = nodes[0, 2]
                if last is None or nodes[-1, 2] > last:
                    last = nodes[-1, 2]
            for time_tick, synth_dict in sorted(type_notes.items()):
                clipboard["notes"].setdefault(time_tick, []).append(synth_dict)
        for _, wall in self.walls.items():
            if first is None or wall[0, 2] < first:
                first = wall[0, 2]
            if last is None or wall[0, 2] > last:
                last = wall[0, 2]
            dest_list, wall_dict = wall_to_synth(self.bpm, wall)
            clipboard[dest_list].append(wall_dict)

        for t in self.lights:
            if first is None or t < first:
                first = t
            if last is None or t > last:
                last = t
            clipboard["lights"].append(round_tick_for_json(t * 64))
        for t in self.effects:
            if first is None or t < first:
                first = t
            if last is None or t > last:
                last = t
            clipboard["effects"].append(round_tick_for_json(t * 64))
        
        if first is None or last is None:
            # completely empty data
            first = last = 0.0

        if realign_start:
            # position of selection start in beats*64 
            clipboard["startMeasure"] = round_tick_for_json(first * 64)
            # position of selection start in ms
            clipboard["startTime"] = beat_to_second(first, self.bpm) * 1000
            # length of the selection in milliseconds
            # and yes, the editor has a typo, so we need to missspell it too
            clipboard["lenght"] = beat_to_second(last, self.bpm) * 1000
        # always update length
        clipboard["lenght"] = beat_to_second(last - first, self.bpm) * 1000
        return json.dumps(clipboard)

    def change_bpm(self, bpm: float) -> None:
        if not bpm > 0:
            raise ValueError("BPM must be greater than 0")
        if self.bpm != bpm:
            ratio = bpm/self.bpm
            self.apply_for_all(movement.scale, [1,1,ratio])
            self.bpm = bpm

@dataclasses.dataclass
class ClipboardDataContainer(DataContainer):
    original_json: str = ""
    selection_length: float = 0.0

    @classmethod
    def from_json(cls, clipboard_json: str, use_original: bool=False) -> "ClipboardDataContainer":
        clipboard = json.loads(clipboard_json)
        if not isinstance(clipboard, dict):
            raise ValueError("clipboard did not contain json dict")
        if "original_json" in clipboard:
            original_json = clipboard["original_json"]
            if use_original:
                clipboard = json.loads(original_json)
        else:
            original_json = clipboard_json
        bpm = clipboard["BPM"]
        startMeasure = clipboard["startMeasure"]
        # r, l, s, b
        right: SINGLE_COLOR_NOTES = {}
        left: SINGLE_COLOR_NOTES = {}
        single: SINGLE_COLOR_NOTES = {}
        both: SINGLE_COLOR_NOTES = {}
        notes_lookup: tuple[SINGLE_COLOR_NOTES, SINGLE_COLOR_NOTES, SINGLE_COLOR_NOTES, SINGLE_COLOR_NOTES] = (right, left, single, both)
        for time_index, time_notes in clipboard["notes"].items():
            for note in time_notes:
                note_type, nodes = note_from_synth(bpm, startMeasure, note)
                notes_lookup[note_type][nodes[0,2]] = nodes

        walls: dict[float, list["numpy array (1, 5)"]] = {}
        # slides (right, left, angle_right, center, angle_right)
        for wall_dict in clipboard["slides"]:
            wall = wall_from_synth(bpm, startMeasure, wall_dict, wall_dict["slideType"])
            walls[wall[0, 2]] = wall
        # other (crouch, square, triangle)
        for wall_type in ("crouch", "square", "triangle"):
            if wall_type + "s" not in clipboard:
                beta_warning()
                continue  # these are only in the beta editor
            for wall_dict in clipboard[wall_type + "s"]:
                wall = wall_from_synth(bpm, startMeasure, wall_dict, WALL_TYPES[wall_type][0])
                walls[wall[0, 2]] = wall
        
        lights = {
            b: np.array([[0,0,b]])
            for t in clipboard["lights"]
            for b in [round_time_to_fractions((t-startMeasure)/64)]
        }
        effects = {
            b: np.array([[0,0,b]])
            for t in clipboard["effects"]
            for b in [round_time_to_fractions((t-startMeasure)/64)]
        }

        return cls(
            # DataContainer
            bpm=bpm, right=right, left=left, single=single, both=both, walls=walls, lights=lights, effects=effects,
            # ClipboardDataContainer
            original_json=original_json, selection_length=second_to_beat(clipboard["lenght"]/1000, bpm),
        )

# Hacky classes used to trick ZipFile._open_to_write
class _TrueInt(int):
    # this makes it think there are already flags, so it doesn't add "permissions: ?rw-------"
    def __bool__(self):
        return True
class _NonAsciiStr(str):
    # this makes it think the filename is not ascii, so it adds the UTF8 flag
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes:
        if encoding == "ascii":
            raise UnicodeEncodeError(encoding, self, 0, 0, "dummy error to trick ZipFile._open_to_write")
        return super().encode(encoding=encoding, errors=errors)

def encode_extra_pkware_win32(timestamp: datetime.datetime) -> bytes:
    # https://mdfs.net/Docs/Comp/Archiving/Zip/ExtraField -> 0x000A
    # encode NTFS timestamp
    ntfs_epoch = datetime.datetime(1601,1,1, 0,0,0,0, tzinfo=datetime.timezone.utc)
    ts_bytes = int((timestamp - ntfs_epoch).total_seconds()*1e7).to_bytes(8, "little")
    
    return struct.pack(
        "<HHIHH",
        0x000a,  # PKWARE Win95/WinNT Extra Field 
        32,      # extra len: 4+2+2+24
        0,       # reserved
        1,       # Tag 1
        24,      # 3*8
    ) + (ts_bytes * 3)

@dataclasses.dataclass
class SynthFileMeta:
    name: str
    artist: str
    mapper: str
    audio_name: str
    explicit: bool = False
    cover_name: str = "No cover"
    cover_data: bytes = b""  # from base64 in json, overwrites separate file
    custom_difficulty_name: str = "Custom"
    custom_difficulty_speed: float = 1.0

    def get_safe_name(self, audio_data: bytes) -> str:
        # editor tracks spectrum data by filename, so just append sha256 of content to filename
        sha256_hash = sha256(audio_data).hexdigest()[:8]
        # <name>_<hash:8 to 64 hexdigits>.<ext> or <name>.<ext>
        m = re.match(r"(.*?)(?:_[0-9a-fA-F]{8,64})*\.([^.]*)$", self.audio_name)
        if m:
            name, ext = m[1], m[2]
        else:  # should only happen if there is no dot in the filename
            name = self.audio_name
            ext = "ogg"
        # cap name to 16 chars, so the filename is not longer than 16+1+8+4=29 chars
        name = "".join(filter(str.isidentifier, name))
        return f"{name[:16]}_{sha256_hash}.{ext}"

@dataclasses.dataclass
class SynthFile:
    meta: SynthFileMeta
    audio: AudioData
    bookmarks: dict[float, str]
    difficulties: dict[str, DataContainer]

    errors: dict[str, list[tuple[JSONParseError, str]]]
    offset_ms: int

    @staticmethod
    def empty_from_audio(audio_file: Union[Path, BytesIO], *, filename: str|None = None, name: str|None = None, artist: str = "unknown artist", mapper: str = "your name here") -> "SynthFile":
        raw_data = audio_file.read_bytes() if isinstance(audio_file, Path) else audio_file.read()
        audio = AudioData.from_raw(raw_data, allow_conversion=True)
        if filename is None:
            if isinstance(audio_file, Path):
                filename = audio_file.stem
            elif name is not None:
                filename = name
            else:
                filename = "unnamed"
        elif "." in filename:
            filename = filename.rsplit(".", 1)[0]
        if name is None:
            name = filename
        return SynthFile(
            meta=SynthFileMeta(
                name=name,
                artist=artist,
                mapper=mapper,
                audio_name=filename + ".ogg",
            ),
            audio=audio,
            bookmarks={},
            difficulties={DIFFICULTIES[0]: DataContainer()},
            errors={},
            offset_ms=0,
        )

    @property
    def bpm(self) -> float:
        for d, c in self.difficulties.items():
            return c.bpm
        return 0.0  # no difficulties

    @staticmethod
    def from_synth(synth_file: Union[Path, BytesIO]) -> "SynthFile":
        errors: dict[str, list[tuple[JSONParseError, str]]] = {}
        with zipfile.ZipFile(synth_file) as inzip:
            # load beatmap json
            beatmap = json.loads(inzip.read(BEATMAP_JSON_FILE))
            audio = AudioData.from_raw(inzip.read(beatmap["AudioName"]))

        bpm: float = beatmap["BPM"]
        difficulties: dict[str, DataContainer] = {}
        for diff in DIFFICULTIES:
            diff_removed : list[tuple[JSONParseError, str]] = []
            # r, l, s, b
            right: SINGLE_COLOR_NOTES = {}
            left: SINGLE_COLOR_NOTES = {}
            single: SINGLE_COLOR_NOTES = {}
            both: SINGLE_COLOR_NOTES = {}
            notes_lookup: tuple[SINGLE_COLOR_NOTES, SINGLE_COLOR_NOTES, SINGLE_COLOR_NOTES, SINGLE_COLOR_NOTES] = (right, left, single, both)
            for time, time_notes in beatmap["Track"][diff].items():
                time_types = set()
                for note in time_notes:
                    try:
                        note_type, nodes = note_from_synth(bpm, 0, note)
                        if note_type in time_types:
                            raise JSONParseError(note, "duplicate note type")
                        time_types.add(note_type)
                        notes_lookup[note_type][nodes[0,2]] = nodes
                    except JSONParseError as jpe:
                        try:
                            time = f"{time} (measure {float(time)/64})"
                        except:
                            pass
                        diff_removed.append((jpe, time))

            walls: dict[float, list["numpy array (1, 5)"]] = {}
            # slides (right, left, angle_right, center, angle_right)
            for wall_dict in beatmap["Slides"][diff]:
                try:
                    wall = wall_from_synth(bpm, 0, wall_dict, wall_dict["slideType"])
                    walls[wall[0, 2]] = wall
                except JSONParseError as jpe:
                    time = str(wall_dict.get("time"))
                    try:
                        time = f"{time} (measure {float(time)/64})"
                    except:
                        pass
                    diff_removed.append((jpe, time))
            # other (crouch, square, triangle)
            for wall_type in ("Crouch", "Square", "Triangle"):
                if wall_type + "s" not in beatmap:
                    beta_warning()
                    continue  # these are only in the beta editor
                for wall_dict in beatmap[wall_type + "s"][diff]:
                    try:
                        wall = wall_from_synth(bpm, 0, wall_dict, WALL_TYPES[wall_type.lower()][0])
                        walls[wall[0, 2]] = wall
                    except JSONParseError as jpe:
                        time = str(wall_dict.get("time"))
                        try:
                            time = f"{time} (measure {float(time)/64})"
                        except:
                            pass
                        diff_removed.append((jpe, time))
            lights = {
                b: np.array([[0,0,b]])
                for t in beatmap["Lights"][diff]
                for b in [round_time_to_fractions(t / 64)]
            }
            effects = {
                b: np.array([[0,0,b]])
                for t in beatmap["Effects"][diff]
                for b in [round_time_to_fractions(t / 64)]
            }
            # add difficulty when there is a note of any type, a wall or lights/effects
            if any(notes_lookup) or walls or lights or effects:
                difficulties[diff] = DataContainer(bpm=bpm, right=right, left=left, single=single, both=both, walls=walls, lights=lights, effects=effects)

            if diff_removed:
                errors[diff] = diff_removed

        if not difficulties:
            # at add least one difficulty, such that BPM can be tracked
            difficulties[DIFFICULTIES[0]] = DataContainer(bpm=bpm)
            
        return SynthFile(
            meta=SynthFileMeta(
                name = beatmap["Name"],
                artist = beatmap["Author"],
                mapper = beatmap["Beatmapper"],
                explicit = beatmap.get("Explicit", False),
                cover_name = Path(beatmap["Artwork"]).stem + ".png",  # change to .png, regardless of input type
                cover_data = base64.b64decode(beatmap["ArtworkBytes"] or b""),  # this seems to be a converted form (to png)
                audio_name = beatmap["AudioName"],
                custom_difficulty_name = beatmap["CustomDifficultyName"],
                custom_difficulty_speed = beatmap["CustomDifficultySpeed"],
            ),
            audio=audio,
            bookmarks={
                # bookmarks are stored in steps of 64 per beat (regardless of BPM)
                round_time_to_fractions(bookmark_dict["time"] / 64): bookmark_dict["name"]
                for bookmark_dict in beatmap["Bookmarks"]["BookmarksList"]
            },
            difficulties=difficulties,
            errors=errors,
            offset_ms=beatmap["Offset"],
        )

    def save_as(self, output_file: Union[Path, BytesIO]) -> None:
        out_buffer = output_file if isinstance(output_file, BytesIO) else BytesIO()  # buffer output zip file in memory, only write on success

        bookmark_list = [
            {"time": round_tick_for_json(t * 64), "name": n}
            for t, n in sorted(self.bookmarks.items())
        ]
        safe_audio_name = self.meta.get_safe_name(self.audio.raw_data)
        now = datetime.datetime.now(datetime.timezone.utc)
        out_beatmap: dict[str, Any] = {
            "Name": self.meta.name,
            "Author": self.meta.artist,
            "Artwork": self.meta.cover_name,
            "ArtworkBytes": base64.b64encode(self.meta.cover_data).decode(),
            "AudioName": safe_audio_name,
            "AudioData": None,
            "AudioFrecuency": self.audio.sample_rate,
            "AudioChannels": self.audio.channels,
            "BPM": self.bpm,
            "Offset": self.offset_ms,
            "Track": {d: {} for d in DIFFICULTIES},  # notes/rails
            "Effects": {d: [] for d in DIFFICULTIES},
            "Bookmarks" : {"BookmarksList": bookmark_list},
            # note: for some of the following I just use dummy value, no idea what they are used for
            "Jumps": {d: [] for d in DIFFICULTIES},  # ?
            "Crouchs": {d: [] for d in DIFFICULTIES},
            "Slides": {d: [] for d in DIFFICULTIES},
            "Lights": {d: [] for d in DIFFICULTIES},
            "Squares": {d: [] for d in DIFFICULTIES},
            "Triangles": {d: [] for d in DIFFICULTIES},
            "DrumSamples": None,  # ?
            "FilePath": "Redacted",
            "IsAdminOnly": False,  # ?
            "EditorVersion": f"SMH v{__version__}",
            "Beatmapper": self.meta.mapper,
            "CustomDifficultyName": self.meta.custom_difficulty_name,
            "CustomDifficultySpeed": self.meta.custom_difficulty_speed,
            "UsingBeatMeasure": True,  # ?
            "UpdatedWithMovementPositions": True,  # ?
            "ProductionMode": False,  # ?
            "Tags": [],  # ?
            "BeatConverted": False,  # ?
            "BeatModified": False,  # ?
            "ModifiedTime": int(now.timestamp()),
            "Explicit": self.meta.explicit,
        }

        # fill per-difficulty arrays/dicts
        for diff, data in self.difficulties.items():
            new_notes: dict[float, list[dict]] = {}
            for note_type, notes in enumerate((data.right, data.left, data.single, data.both)):
                type_notes: dict[float, dict] = {}  # buffer single type to avoid multiple of the same type at the same time
                for time_index, nodes in sorted(notes.items()):
                    type_notes[round_tick_for_json_index(time_index * 64)] = note_to_synth(data.bpm, note_type, nodes)
                for time_tick, synth_dict in sorted(type_notes.items()):
                    existing = new_notes.setdefault(time_tick, [])
                    # edit by reference
                    existing.append(
                        # keep same order as editor
                        {"Id": f"Note_{time_tick}{NOTE_TYPE_STRINGS[note_type]}{len(existing)}", "ComboId": -1}
                        | synth_dict
                        | {"Direction": 0}
                    )
            out_beatmap["Track"][diff] = {k: v for k,v in sorted(new_notes.items())}
            walls: dict[str, list[dict]] = {
                "crouchs": [],
                "squares": [],
                "triangles": [],
                "slides": [],
            }
            for _, wall in sorted(data.walls.items()):
                dest_list, wall_dict = wall_to_synth(data.bpm, wall)
                walls[dest_list].append(wall_dict)
            for t, wall_list in walls.items():
                out_beatmap[t.capitalize()][diff] = wall_list
            out_beatmap["Lights"][diff] = [round_tick_for_json(t * 64) for t in data.lights]
            out_beatmap["Effects"][diff] = [round_tick_for_json(t * 64) for t in data.effects]

        # write modified beatmap json, in a way that closely mirrors the editor
        beatmap_json = json.dumps(out_beatmap, indent=2, allow_nan=False)
        meta_json = json.dumps({
            "name": self.meta.name,
            "artist": self.meta.artist,
            "duration": f"{self.audio.duration//60:02.0f}:{self.audio.duration%60:02.0f}",
            "coverImage": self.meta.cover_name,
            "audioFile": safe_audio_name,
            "supportedDifficulties": [
                d if self.difficulties.get(d, False) else "" for d in DIFFICULTIES
            ],
            "bpm": self.bpm,
            "mapper": self.meta.mapper,
        }, indent=4, allow_nan=False)
        
        with zipfile.ZipFile(out_buffer, "w") as outzip:
            def make_fileinfo(filename: str) -> zipfile.ZipInfo:
                info = zipfile.ZipInfo(filename, now.timetuple()[:6])
                info.compress_type = zipfile.ZIP_DEFLATED
                # fake various file headers
                info.create_system = 0
                info.create_version = zipfile.ZIP64_VERSION
                # editor files also contain NTFS timestamps
                info.extra = encode_extra_pkware_win32(now)

                # trick header encoder
                info.external_attr = _TrueInt(0)
                info.filename = _NonAsciiStr(filename)
                return info
            
            outzip.writestr(make_fileinfo(BEATMAP_JSON_FILE), codecs.BOM_UTF8 + beatmap_json.encode("utf-8").replace(b"\n", b"\r\n"))
            outzip.writestr(make_fileinfo(safe_audio_name), self.audio.raw_data)
            outzip.writestr(make_fileinfo(self.meta.cover_name), self.meta.cover_data)
            outzip.writestr(make_fileinfo(METADATA_JSON_FILE), codecs.BOM_UTF16_LE + meta_json.encode("utf-16-le"))
        # write output zip
        if isinstance(output_file, BytesIO):
            output_file.seek(0)
        else:
            output_file.write_bytes(out_buffer.getbuffer())

    def reload(self, new_file: Union[Path, BytesIO]) -> "SynthFile":
        new = type(self).from_synth(new_file)
        if new.audio.raw_data == self.audio.raw_data:
            # keep audio cache, if applicable
            new.audio = self.audio
        return new

    def change_bpm(self, bpm: float) -> None:
        if not bpm > 0:
            raise ValueError("BPM must be greater than 0")
        if self.bpm != bpm:
            ratio = bpm/self.bpm
            self.bookmarks = {
                time * ratio: name
                for time, name in self.bookmarks.items()
            }
            for c in self.difficulties.values():
                c.change_bpm(bpm)

    def offset_everything(self, delta_s: float|None = None, delta_b: float|None = None) -> None:
        if (delta_s is not None) == (delta_b is not None):
            raise ValueError("Specify either delta_s or delta_b")
        elif delta_b is None and delta_s is not None:
            delta_b = second_to_beat(delta_s, self.bpm)
        if delta_b:
            self.bookmarks = {
                time + delta_b: name
                for time, name in self.bookmarks.items()
            }
            for c in self.difficulties.values():
                c.apply_for_all(movement.offset, [0,0,delta_b])

    def change_offset(self, offset_ms: int) -> None:
        if offset_ms < 0:
            raise ValueError("Offset must be greater than 0")
        if self.offset_ms != offset_ms:
            self.offset_everything(delta_s=(offset_ms-self.offset_ms)/1000)
            self.offset_ms = offset_ms

    def merge(self, other: "SynthFile", adjust_bpm: bool = True, merge_bookmarks: bool = True) -> None:
        if adjust_bpm and self.bpm != other.bpm:
            other.change_bpm(self.bpm)
        if adjust_bpm and self.offset_ms != other.offset_ms:
            other.change_offset(self.offset_ms)
        if merge_bookmarks:
            self.bookmarks |= other.bookmarks
        for d, c in other.difficulties.items():
            if d in self.difficulties:
                self.difficulties[d].merge(c)
            else:
                self.difficulties[d] = c
        for d, e in other.errors.items():
            if d in self.errors:
                self.errors[d].extend(e)
            else:
                self.errors[d] = e

    def with_added_silence(self, *, before_start_ms: int|None = 0, after_end_ms: int|None = 0) -> "SynthFile":
        if not before_start_ms and not after_end_ms:
            return
        if before_start_ms is None:
            before_start_ms = 0
        if after_end_ms is None:
            after_end_ms = 0
        new_audio = self.audio.with_silence(before_start_s=before_start_ms/1000, after_end_s=after_end_ms/1000)
        output = dataclasses.replace(self, audio=new_audio)
        if before_start_ms:
            output.offset_everything(delta_s=before_start_ms/1000)
        return output

# convenience wrappers

# clipboard json
def import_clipboard_json(original_json: str, use_original: bool = False) -> ClipboardDataContainer:
    return ClipboardDataContainer.from_json(clipboard_json=original_json, use_original=use_original)

def export_clipboard_json(data: DataContainer, realign_start: bool = True) -> str:
    return data.to_clipboard_json(realign_start=realign_start)

# direct clipboard
def import_clipboard(use_original: bool = False) -> ClipboardDataContainer:
    return import_clipboard_json(pyperclip.paste(), use_original)

def export_clipboard(data: DataContainer, realign_start: bool = True):
    pyperclip.copy(export_clipboard_json(data, realign_start))

@contextmanager
def clipboard_data(use_original: bool = False, realign_start: bool = True) -> Generator[ClipboardDataContainer, None, None]:
    # Usage:
    #   with synth_format.clipboard_data() as data:
    #     data.apply_for_all(...)
    data = ClipboardDataContainer.from_json(pyperclip.paste(), use_original=use_original)
    yield data
    pyperclip.copy(data.to_clipboard_json(realign_start=realign_start))

# file
def import_file(file_path: Union[Path, BytesIO]) -> SynthFile:
    return SynthFile.from_synth(file_path)

@contextmanager
def file_data(filename: str|Path|None = None, save_suffix: str|None = "_out") -> Generator[SynthFile, None, None]:
    if filename is None and len(sys.argv) == 2 and sys.argv[1].endswith(".synth") and Path(sys.argv[1]).is_file():
        # if given a single command line argument that is a .synth file, use that
        fp = Path(sys.argv[1])
    elif filename is not None:
        fp = Path(filename)
    else:
        raise ValueError("No filename provided")
    print(f"Loading {fp.absolute()}")
    f = SynthFile.from_synth(fp)
    yield f
    if save_suffix is not None:
        fp_out = fp.with_stem(f"{fp.stem}{save_suffix}")
        print(f"Saving {fp_out.absolute()}")
        f.save_as(fp_out)
